library geiger_replication;

import 'dart:io';
import 'dart:async';
import 'dart:convert';
import 'dart:developer';

// ignore: library_prefixes
import 'package:encrypt/encrypt.dart' as Enc;
import 'package:uuid/uuid.dart';
import 'package:intl/intl.dart';
import 'package:intl/locale.dart';

import 'package:cloud_adapter/src/cloud_models/recommendation.dart';
import 'package:cloud_adapter/src/cloud_models/security_defender_user.dart';
import 'package:cloud_adapter/src/cloud_models/security_defenders_organization.dart';
import 'package:cloud_adapter/src/cloud_models/user.dart';
import '../cloud_models/cloud_connected_tool_status_recommendation.dart';
import '../cloud_models/cloud_connected_tool_recommendation.dart';
import '../cloud_models/threats.dart';
import '../cloud_models/event.dart';
import '../cloud_models/company_event.dart';
import '../cloud_models/short_user.dart';
import 'package:cloud_adapter/src/replication_exception.dart';
import 'package:cloud_adapter/src/cloud_models/threat_dict.dart';
import 'package:cloud_adapter/src/cloud_models/threat_weights.dart';
import 'package:cloud_adapter/src/service/cloud_service/cloud_service.dart';

import 'package:geiger_localstorage/geiger_localstorage.dart' as toolbox_api;
import 'package:geiger_api/geiger_api.dart';

import 'node_listener.dart';
import 'replication_message_listener.dart';

import '../replication_controller.dart';

class ReplicationService implements ReplicationController {
  /// Cloud Service
  /// Enables Replication & Pairing
  late CloudService cloud;

  /// GEIGER API
  late final GeigerApi _localMaster;

  /// Storage controller
  late final toolbox_api.StorageController _storageController;

  /// Storage Listener
  // ignore: unused_field
  late final toolbox_api.StorageListener _storageListener;

  /// Message Listener
  ReplicationMessageListener? _messageListener;
  bool _isPluginListenerRegistered = false;
  // ignore: prefer_final_fields
  List<MessageType> _handledEvents = [];

  late final String _username, _company, _currentDevice;
  bool usernameInitialized = false;
  bool companyInitialized = false;
  bool currentDeviceInitialized = false;

  final String _devicesPairing = 'devicesPairing';
  final String _employeesPairing = 'employeesPairing';
  final String _replicationAPI = 'replication_plugin';
  final String __debugLine = '[REPLICATION SERVICE] ';
  final String _enterpriseUsers = ":Enterprise:Users";
  final String _geigerScoreAggNodeName = "geigerscoreaggregate";
  final String _sensorCloudConnectedTLP = 'white';
  final String _metrics = 'metrics';
  final String _recommendations = 'recommendations';

  // ignore: unused_field
  final List<String> _rootDoNotReplicate = [
    ':Keys',
    ':Users',
    ':Enterprise',
    ':Global',
    ':Local'
  ];

  final List<String> _typesOfSensors = [
    'sensorCloudConnected',
    'sensorRecommendation',
    'sensorStatus'
  ];

  final List<String> _typesOfTrainings = [
    'trainingCloudConnected',
    'trainingRecommendation',
    'trainingStatus'
  ];

  //MAP OF CLOUD EVENTS TO JSON
  List<Map<dynamic, dynamic>> cloudNodeList = [];
  //STRING LIST OF THE PATH THAT COVERS EACH EVENT
  List<String> cloudNodePath = [];
  //USER EVENT DATA
  List<Event> cloudEvents = [];
  //STRING LIST OF CLOUD EVENTS
  List<String> userEvents = [];
  //STRING OF CONTENTS OF EVENTS
  List<String> contentEvents = [];
  //STRING LIST OF COMPANY EVENTS;
  List<String> companyEvents = [];

  String sharedScore = '', username = '', sharedScoreDate = '';

  ReplicationService() {
    cloud = CloudService();
  }

  @override
  Future<bool> checkConnection() async {
    log(__debugLine + 'checkConnection');
    try {
      /// CONNECTION WITH THE CLOUD
      /// PROVIDE FINAL CLOUD URL
      final response = await InternetAddress.lookup('37.48.101.252');
      if (response.isNotEmpty) {
        return true;
      } else {
        return false;
      }
    } on SocketException catch (e) {
      log(__debugLine + 'EXCEPTION: ' + e.toString());
      return false;
    }
  }

  @override
  Future<bool> checkReplication() async {
    log(__debugLine + 'checkReplication');
    try {
      // ignore: unused_local_variable
      toolbox_api.Node checker =
          await getNode(':Local:Replication:LastReplication');
      return true;
    } catch (e) {
      log(__debugLine + 'EXCEPTION: ' + e.toString());
      return false;
    }
  }

  @override
  Future<bool> checkReplicationTimingStrategy() async {
    log(__debugLine + 'checkReplicationTimingStrategy');
    try {
      // ignore: unused_local_variable
      toolbox_api.Node checker =
          await getNode(':Local:Replication:LastReplication');
      DateTime _actual = DateTime.now();
      DateTime _fromDate = _actual.subtract(Duration(days: 30));
      int milSEpoch = _fromDate.millisecondsSinceEpoch;
      int nodeTimestamp = checker.lastModified;
      return milSEpoch < nodeTimestamp ? false : true;
    } catch (e) {
      log(__debugLine + 'EXCEPTION');
      return false;
    }
  }

  @override
  Future<bool> geigerReplicationUpdate() async {
    /// Follow diagram
    /// 3 steps replication
    /// Cloud to device
    /// Device to cloud
    /// Storage listener
    /// Take care of strategies

    /// 1. INIT STORAGE
    /// this is done with initgeigerstoremethod providing an already initiated storage controller
    cloud = await CloudService.create();

    /// CHECK WHEN LAST REPLICATION TOOK PLACE
    /// Nodes that handle replication:
    /// :Replication:LastReplication
    DateTime _actual = DateTime.now();
    bool _fullRep;

    /// 180 -> Default expiring date
    DateTime _fromDate = _actual.subtract(Duration(days: 180));
    log('[1st FLOW] - TIMESTAMP CHECKER');
    try {
      toolbox_api.Node timeChecker =
          await getNode(':Local:Replication:LastReplication');
      DateTime _lastTimestamp =
          DateTime.fromMillisecondsSinceEpoch(timeChecker.lastModified);
      Duration _diff = _actual.difference(_lastTimestamp);
      if (_diff.inDays > 30) {
        /// FULL REPLICATION TAKES PLACE
        log("FULL REPLICATION");
        _fullRep = true;
      } else {
        /// PARTIAL REPLICATION TAKES PLACE
        log("PARTIAL REPLICATION");
        _fullRep = false;
        _fromDate = _lastTimestamp;
      }
    } catch (e) {
      log('NO REPLICATION NODE FOUND - NO REPLICATION HAS BEEN DONE BEFORE');

      /// FULL REPLICATION TAKES PLACE
      log("FIRST FULL REPLICATION");
      _fullRep = true;
    }

    // 2. GET USER DATA - CHECK USER IN CLOUD
    log('[1st FLOW] - CHECK USER DATA');
    try {
      await getCurrentUserUUID();
      //CHECK IF USER ALREADY IN CLOUD & IF NOT, CREATE ONE
      bool exists = await cloud.userExists(_username);
      if (exists == false) {
        toolbox_api.Node localUser =
            await getNode(':Enterprise:Users:$_username');
        String name = await localUser
            .getValue('name')
            .then((value) => value!.getValue("en").toString());
        toolbox_api.Node keys = await getNode(':Keys');
        final nodeKey = Enc.Key.fromSecureRandom(32).base64.toString();
        keys.addOrUpdateValue(toolbox_api.NodeValueImpl('publicKey', nodeKey));
        await _storageController.addOrUpdate(keys);
        String key = await keys
            .getValue('publicKey')
            .then((value) => value!.getValue("en").toString());

        toolbox_api.Node enterprise = await getNode(':Enterprise');
        String email = await enterprise
            .getValue('email')
            .then((value) => value!.getValue("en").toString());

        await cloud.createUser(_username, email, null, null, name, key);

        /// FULL REPLICATION TAKES PLACE
        _fullRep = true;
      }
    } catch (e) {
      toolbox_api.StorageException(
          'USER DATA NOT FOUND. ERROR: ' + e.toString());
      return false;
    }

    /// CLOUD TO LOCAL REPLICATION
    log("[2nd FLOW] - CLOUD TO LOCAL");
    await cloudToLocalReplication(_username, _actual, _fromDate, _fullRep);

    /// LOCAL TO CLOUD REPLICATION
    log("[3rd FLOW] - LOCAL TO CLOUD");
    await localToCloudReplication(_username, _actual, _fromDate, _fullRep);

    /// UPDATE WHEN LAST REPLICATION TOOK PLACE
    log("[4th FLOW] - UPDATE WHEN LAST REPLICATION TOOK PLACE");
    await updateReplicationNode(_actual);

    return true;
  }

  @override
  Future<bool> geigerReplication(
      deleteHandler, createHandler, updateHandler, renameHandler) async {
    log(__debugLine + 'STARTING GEIGER REPLICATION REGULAR WITH GLOBAL');
    final stopWatch = Stopwatch();
    stopWatch.start();
    await geigerReplicationUpdate();

    log("keys dump ==> ${await _storageController.dump(":Keys")}");

    /// STORAGE LISTENER
    log("[5th FLOW] - LISTENER REPLICATION");
    await geigerReplicationListener(
        deleteHandler, createHandler, updateHandler, renameHandler);
    log("EXECUTION TIME OF GEIGER REPLICATION : " +
        stopWatch.elapsed.toString());
    stopWatch.stop();
    return true;
  }

  @override
  Future<bool> updateGlobalData() async {
    bool result = false;
    try {
      /// Updates GEIGER Defined Threats
      await updateThreats();

      /// Updates Threat Weights
      await updateThreatWeights();

      /// Updates Global Recommendations in a Cloud to Local way
      await updateRecommendations();

      /// Updates Security Defenders Info
      await updateSecurityDefendersInfo();

      /// Get sensor events Information
      await getSensorData();

      /// Get training events Information
      await getTrainingData();

      result = true;
    } catch (e) {
      log("[UPDATE GLOBAL DATA EXCEPTION]: " + e.toString());
    }
    return result;
  }

  Future<void> cloudToLocalReplication(String _username, DateTime _actual,
      DateTime _fromDate, bool _fullRep) async {
    log("CLOUD TO LOCAL REPLICATION - Function cloudToLocalReplication");

    /// WITH DATETIME TAKE EVENTS FROM THE CLOUD
    List<String> events;

    // FILTER BY DATE
    if (_fullRep == true) {
      events = await cloud.getUserEvents(_username);
    } else {
      var filter =
          DateFormat("yyyy-MM-dd'T'hh:mm:ss.SSSS'Z'").format(_fromDate);
      events =
          await cloud.getUserEventsDateFilter(_username, filter.toString());
    }

    List<Event> waitingEvents = [];
    for (var userEvent in events) {
      Event singleOne = await cloud.getSingleUserEvent(_username, userEvent);
      Map<dynamic, dynamic> data = getPartialJsonString(singleOne.content!);

      // HAVE A VAR WITH DATA
      if (userEvents.contains(userEvent) == false) {
        userEvents.add(userEvent);
        cloudEvents.add(singleOne);
        cloudNodeList.add(data);
        cloudNodePath.add(data['path']);
      } else {
        int index = userEvents.indexOf(userEvent);
        Event toCompare = cloudEvents[index];
        DateTime? evToCompare = toCompare.last_modified;
        DateTime? newTime = singleOne.last_modified;
        //AS IN CLOUD, COMPARE DATETIME
        Duration _diff = newTime!.difference(evToCompare!);
        if (_diff.inDays > 0) {
          userEvents[index] = userEvent;
          cloudNodeList[index] = (data);
          cloudNodePath[index] = (data['path']);
          cloudEvents[index] = singleOne;
        }
      }

      String type = singleOne.type.toString();
      //String tlp = singleOne.tlp.toString();
      String id = singleOne.id_event.toString();
      String owner = singleOne.owner.toString();
      if (owner != _username) {
        if (type.toLowerCase() == "keyvalue") {
          log(
              'CLOUD TO LOCAL REPLICATION - NO E2EE - FUNCTION cloudToLocalReplication');
          toolbox_api.Node node = convertJsonStringToNode(singleOne.content!);
          await updateLocalNodeWithCloudNode(node, _username);
        }
        if (type.toLowerCase() == "user_object") {
          log(
              'CLOUD TO LOCAL REPLICATION - E2EE - FUNCTION cloudToLocalReplication');
          try {
            Map<dynamic, dynamic> jsonify = await decryptCloudData(singleOne);
            singleOne.setContent = jsonEncode(jsonify);
            toolbox_api.Node node = convertJsonStringToNode(singleOne.content!);
            await checkParents(node.path);
            await updateLocalNodeWithCloudNode(node, _username);
          } catch (e) {
            log(
                "ERROR DURING CLOUD TO LOCAL REPLICATION E2EE:" + e.toString());

            /// KAY MAY NOT BE REPLICATED YET - CREATE NEW LIST AND DO SAME OPERATION
            waitingEvents.add(singleOne);
            log(
                'FAILURE GETTING KEYS FOR NODE - FUNCTION cloudToLocalReplication ID-EVENT: $id');
          }

          /*else{
            try {
              log(
                  'CLOUD TO LOCAL REPLICATION - NO ENCRYPTION ENABLED - FUNCTION cloudToLocalReplication');
              toolbox_api.Node node = convertJsonStringToNode(
                  singleOne.content!);
              await checkParents(node.path);
              await updateLocalNodeWithCloudNode(node, _username);
            } catch (e){
              log("ERROR WITHIN FUNCTION cloudToLocalReplication LINE 328: "+e.toString());
              waitingEvents.add(singleOne);
            }
          }*/
        }
      }
    }

    log("WAITING EVENTS - FUNCTION cloudToLocalReplication");
    for (var waiting in waitingEvents) {
      String waitingId = waiting.id_event.toString();
      try {
        ///CREATE A SEARCH CRITERIA TO FIND KEY WITH GIVEN PATH
        ///ELSE CREATE A NODE WITH VISIBILITY AMBER
        ///AND REPLICATE
        Map<dynamic, dynamic> jsonify = await decryptCloudData(waiting);
        waiting.setContent = jsonEncode(jsonify);
        toolbox_api.Node node = convertJsonStringToNode(waiting.content!);
        await checkParents(node.path);
        await updateLocalNodeWithCloudNode(node, _username);
      } catch (e) {
        log(e.toString());
        log(
            'FAILURE GETTING KEYS FOR NODE - FUNCTION cloudToLocalReplication: $waitingId');
        try {
          toolbox_api.Node node = convertJsonStringToNode(waiting.content!);
          await updateLocalNodeWithCloudNode(node, _username);
        } catch (e) {
          log(
              "GET SHARED NODES FAILURE - FUNCTION cloudToLocalReplication: " +
                  e.toString());
        }
      }
    }

    await updateGlobalData();
  }

  Future<void> localToCloudReplication(String _username, DateTime _actual,
      DateTime _fromDate, bool _fullRep) async {
    log(
        "REPLICATION - LOCAL TO CLOUD NODES REPLICATION - FUNCTION localToCloudReplication");

    /// START OF THE SECOND DIAGRAM
    /// START CLEANING DATA AND REMOVING TOMBSTONES
    ///await cleanData(_actual);
    List<toolbox_api.Node> nodeList = [];
    if (_fullRep == true) {
      /// FIND ALL THE NODES
      ///nodeList.addAll(await getAllNodes(false));
      nodeList.addAll(await getSpecificNodes());
    } else {
      /// ASK FOR SEARCH CRITERIA NODE BASED
      //log('PARTIAL REPLICATION');
      nodeList.addAll(await getAllNodesLastModified(_fromDate));
    }

    /// SORT NODES BY TIMESTAMP
    nodeList.sort((a, b) => DateTime.fromMillisecondsSinceEpoch(a.lastModified)
        .compareTo(DateTime.fromMillisecondsSinceEpoch(b.lastModified)));

    /// GET ALL EVENTS OF A USER - THE NAME CAN BE ANYTHING
    /// UUID IS RECOMMENDED BUT NOT MANDATORY
    /// CONVERT ALL TO NODES AND CHECK
    List<String> events;
    try {
      events = await cloud.getUserEvents(_username);
      for (var userEvent in events) {
        try {
          Event newCloudEvent =
              await cloud.getSingleUserEvent(_username, userEvent);
          Map<dynamic, dynamic> data =
              getPartialJsonString(newCloudEvent.content!);
          // HAVE A VAR WITH DATA
          if (userEvents.contains(userEvent) == false) {
            userEvents.add(userEvent);
            cloudEvents.add(newCloudEvent);
            cloudNodeList.add(data);
            cloudNodePath.add(data['path']);
          } else {
            /// CHECK TIMESTAMP
            int index = userEvents.indexOf(userEvent);
            Event toCompare = cloudEvents[index];
            DateTime? evToCompare = toCompare.last_modified;
            DateTime? newTime = newCloudEvent.last_modified;
            //AS IN CLOUD, COMPARE DATETIME
            Duration _diff = newTime!.difference(evToCompare!);
            if (_diff.inDays > 0) {
              userEvents[index] = userEvent;
              cloudNodeList[index] = (data);
              cloudNodePath[index] = (data['path']);
              cloudEvents[index] = newCloudEvent;
            }
          }
        } catch (e) {
          log("ERROR GETTING SINGLE USER EVENT: " + e.toString());
        }
      }
    } catch (e) {
      log("ERROR GETTING CLOUD USER EVENTS: " + e.toString());
    }

    if (nodeList.isEmpty == false) {
      for (var sorted in nodeList) {
        /// CHECK TLP
        String tlp = sorted.visibility.toValueString();
        //String identifier = sorted.name.toString();
        String fullPath = sorted.path;

        /// CHECK CONSENT
        /// TBD
        /// 3 AGREEMENTS: NEVER, NO ONCE, AGREE ONCE
        /// CREATE EVENT
        String uuid = await cloud.generateUUID();
        Event toCheck = Event(id_event: uuid, tlp: tlp.toUpperCase());
        toCheck.encoding = 'ascii';
        if (tlp.toLowerCase() != 'black' &&
            sorted.path.startsWith(':Local') == false &&
            sorted.path.startsWith(':Global') == false &&
            sorted.tombstone == false) {
          if (tlp.toLowerCase() == 'red') {
            log("TLP:RED NODE");
            try {
              Map<dynamic, dynamic> convertedNode =
                  await encryptCloudData(sorted);
              toCheck.content = json.encode(convertedNode);
            } catch (e) {
              log(e.toString());
              toCheck.content =
                  json.encode(await convertNodeToJsonString(sorted));
              log('FAILURE GETTING KEYS LOCAL TO CLOUD REPLICATION');
            }
            toCheck.type =
                (sorted.name.toLowerCase() == _geigerScoreAggNodeName)
                    ? 'aggregate_score'
                    : 'user_object';
          } else {
            toCheck.type =
                (sorted.name.toLowerCase() == _geigerScoreAggNodeName)
                    ? 'aggregate_score'
                    : 'user_object';
            toCheck.content =
                json.encode(await convertNodeToJsonString(sorted));
          }
          toCheck.owner = _username;
          //CHECK IF EVENT IN CLOUD
          try {
            // IF OK
            // CHECK TIMESTAMP
            // IF NEEDED -> PUT
            if (cloudNodePath.contains(fullPath) == true) {
              log("NODE ALREADY IN CLOUD");
              int index = cloudNodePath.indexOf(fullPath);
              DateTime toCompare = DateTime.fromMillisecondsSinceEpoch(
                  cloudNodeList[index]['lastModified']);
              DateTime nodeTime =
                  DateTime.fromMillisecondsSinceEpoch(sorted.lastModified);
              //AS IN CLOUD, COMPARE DATETIME
              Duration _diff = nodeTime.difference(toCompare);
              if (_diff.inDays > 0) {
                log("UPDATE LOCAL NODE IN CLOUD");
                await cloud.updateEvent(_username, userEvents[index], toCheck);
              }
            } else {
              log("NODE NOT IN CLOUD -- ADD LOCAL NODE TO CLOUD");
              await cloud.createEvent(_username, toCheck);
              toCheck.last_modified = DateTime.now();
              cloudEvents.add(toCheck);
              //cloudNodeList.add(data);
              //cloudNodePath.add(data['path']);
            }
          } catch (e) {
            log("[EXCEPTION] CATCH ADD LOCAL NODE TO CLOUD");
            log("ERROR check in if the event is in the cloud: " + e.toString());
            // IF NO EVENT IS RETURNED -> POST NEW EVENT
            await cloud.createEvent(_username, toCheck);
          }
        } else {
          log("LOCAL NODE MUST NOT BE REPLICATED: " + sorted.toString());
        }
      }
    }
    log("END LOCAL TO CLOUD REPLICATION");
    //toolbox_api.SearchCriteria criteria = toolbox_api.SearchCriteria();
    //List<toolbox_api.Node> nodeListFirst = await _storageController.search(criteria);
  }

  @override
  Future<bool> checkPairing(
      String qrUserId, String qrDeviceId, String typeOfPairing) async {
    return (typeOfPairing.contains('device'))
        ? await nodeExists(
            '$_enterpriseUsers:$_username:$_devicesPairing:$qrDeviceId')
        : await nodeExists(
            '$_enterpriseUsers:$_username:$_employeesPairing:$qrUserId');
  }

  /// INFO 		 					              --- QR
  /// PAIRING USER 					          --- qrUserId
  /// PAIR DeviceId					          --- qrDeviceId
  /// public key   					          --- publicKey
  /// agreement    					          --- {"in","out","both"}
  /// type                            --- device / employee
  @override
  Future<bool> setPair(String qrUserId, String qrDeviceId, String publicKey,
      String agreement, String type) async {
    log("START SET PAIR METHOD");

    // CREATE PAIRING STRUCTURE
    log('[PAIRING STRUCTURE] CREATE PAIRING');
    await createPairingStructure(qrUserId, qrDeviceId, publicKey, type);

    /// REPLICATES INTO THE CLOUD WITH COMPLEMENTARY AGREEMENTS
    String complementValue;
    if (agreement == "in") {
      complementValue = "out";
    } else if (agreement == "out") {
      complementValue = "in";
    } else if (agreement == "both") {
      complementValue = "both";
    } else {
      log("Not valid agreement value. Choose between: {'in','out','both'}");
      return false;
    }

    /// check if mutual merge exists
    /// if one user agrees in - the other must agree out and vice versa
    /// if agree both - the other must agree both
    /// check userId1 agreements
    await checkMutualMerge(_username, qrUserId, agreement, type);
    await checkMutualMerge(qrUserId, _username, complementValue, type);

    log("all pairedDevices after all => ${await _storageController.dump('$_enterpriseUsers:$_username:$_devicesPairing')}");
    log("all pairedEmployees after all => ${await _storageController.dump('$_enterpriseUsers:$_username:$_employeesPairing')}");
    return true;
  }

  Future<void> checkMutualMerge(
      String userOne, String userTwo, String agreement, String type) async {
    try {
      List<String> agreeUser1 = await cloud.getMergedAccounts(userOne);
      if (agreeUser1.contains(userTwo) == false) {
        log("CREATE MERGE BETWEEN $userOne AND $userTwo");
        await cloud.createMerge(userOne, userTwo, agreement, type);
      } else {
        log("CLOUD AGREEMENT ALREADY EXIST");
      }
    } catch (e) {
      log("ERROR CREATING MERGE REPLICATION SERVICE: " + e.toString());
      log("CREATE MERGE BETWEEN $userOne AND $userTwo");
      await cloud.createMerge(userOne, userTwo, agreement, type);
    }
  }

  Future<bool> deleteMutualMerge(
      String userOne, String userTwo, String typeOfPairing) async {
    try {
      List<String> agreeUser1 = await cloud.getMergedAccounts(userOne);
      if (agreeUser1.isNotEmpty && agreeUser1.contains(userTwo) == true) {
        try {
          log(
              "DELETE MERGE BETWEEN $userOne AND $userTwo TYPE $typeOfPairing");
          await cloud.deleteMerged(userOne, userTwo, typeOfPairing);
          return true;
        } catch (e) {
          log("[CLOUD EXCEPTION] COULD NOT FIND MERGE");
          return false;
        }
      }
    } catch (e) {
      log("[CLOUD EXCEPTION] ERROR GETTING MERGE ACCOUNTS");
      return false;
    }
    return false;
  }

  @override
  Future<bool> updatePair(String userId1) async {
    log("[REPLICATION SERVICE] UPDATE PAIR METHOD");

    /// CHECK THE CLOUD FOR AGREEMENTS
    /*List<String> mergeds = await cloud.getMergedAccounts(userId1);
    if (mergeds.isNotEmpty) {
      for (var user in mergeds) {
        ShortUser u = await cloud.getMergedInfo(userId1, user);

        /// CHECK LOCAL PAIRING NODE
        try {
          // ignore: unused_local_variable
          toolbox_api.Node replicationParent = await getNode(':Local:Pairing');
        } catch (e) {
          log("ERROR CHECKING PARING NODE LINE 649 REPLICATION SERVICE: "+e.toString());
          toolbox_api.Node replicationParent =
              toolbox_api.NodeImpl(':Local:Pairing', _replicationAPI);
          await _storageController.add(replicationParent);
        }
        try {
          // ignore: unused_local_variable
          toolbox_api.Node replicationNode =
              await getNode(':Local:Pairing:$user');
          log("PAIRING ALREADY EXIST");
        } catch (e) {
          toolbox_api.Node replicationNode = toolbox_api.NodeImpl(
              ':Local:Pairing:$user', 'ReplicationService');
          if (u.getPublicKey != null) {
            replicationNode.addOrUpdateValue(
                toolbox_api.NodeValueImpl("key", u.getPublicKey!));
          } else {
            replicationNode
                .addOrUpdateValue(toolbox_api.NodeValueImpl("key", ""));
          }
          await _storageController.add(replicationNode);
        }
        await getSharedNodes(userId1, user);
      }
    } else {
      return false;
    }*/
    return true;
  }

  @override
  Future<bool> unpair(
      String qrUserId, String qrDeviceId, String typeOfPairing) async {
    log("START UNPAIR METHOD");

    /// DELETE MUTUAL MERGE
    bool ok = await deleteMutualMerge(_username, qrUserId, typeOfPairing);
    if (!ok) {
      await deleteMutualMerge(qrUserId, _username, typeOfPairing);
    }

    if (typeOfPairing.contains('device')) {
      bool exist = await nodeExists(
          '$_enterpriseUsers:$_username:$_devicesPairing:$qrDeviceId');
      if (exist) {
        await _storageController.delete(
            '$_enterpriseUsers:$_username:$_devicesPairing:$qrDeviceId');
        log("Local Enterprise After => ${await _storageController.dump('$_enterpriseUsers:$_username:$_devicesPairing')}");
        log("[GEIGER CLOUD] UNPAIR FINISHED OK");
        return true;
      } else {
        log("UNPAIR qrDeviceId NOT FOUND ON LOCAL STORAGE");
        return false;
      }
    } else {
      bool exist = await nodeExists(
          '$_enterpriseUsers:$_username:$_employeesPairing:$qrUserId');
      if (exist) {
        await _storageController.delete(
            '$_enterpriseUsers:$_username:$_employeesPairing:$qrUserId');
        log("Local Enterprise After => ${await _storageController.dump('$_enterpriseUsers:$_username:$_employeesPairing')}");
        log("[GEIGER CLOUD] UNPAIR FINISHED OK");
        return true;
      } else {
        log("UNPAIR qrUserId NOT FOUND ON LOCAL STORAGE");
        return false;
      }
    }
  }

  /// TODO NO NEED FOR THIS shareNode
  @override
  Future<bool> shareNode(
      String nodePath, String senderUserId, String receiverUserId) async {
    /// check local pairing
    /// check cloud pairing
    /// get key
    /// create cloud event e2ee (if public key - else no e2ee)
    /// post event
    try {
      log("SHARE NODE BETWEEN PAIRED DEVICES");
      toolbox_api.Node node = await getNode(':Local:Pairing:$receiverUserId');
      String encryptedKey = await node
          .getValue('key')
          .then((value) => value!.getValue("en").toString());
      String agreeValue = await node
          .getValue('agreement')
          .then((value) => value!.getValue("en").toString());
      if (agreeValue == 'out' || agreeValue == 'both') {
        try {
          log("Agreement valid");
          List<String> agreements = await cloud.getMergedAccounts(senderUserId);
          if (agreements.contains(receiverUserId) == true) {
            ShortUser data =
                await cloud.getMergedInfo(senderUserId, receiverUserId);
            String? publicKey = data.getPublicKey;
            if (publicKey != null) {
              try {
                final keyVal = Enc.Key.fromUtf8(publicKey);
                final iv = Enc.IV.fromLength(16);
                final enc =
                    Enc.Encrypter(Enc.AES(keyVal, mode: Enc.AESMode.cfb64));
                // need to decrypt the encryptedKey
                // DECRYPT DATA
                encryptedKey = enc.decrypt(
                    encryptedKey.split(':')[1] as Enc.Encrypted,
                    iv: iv);
              } catch (e) {
                log("ISSUE DECRYPTING DATA");
              }
            } else {
              encryptedKey = encryptedKey.split(':')[1];
            }
            // node to send
            toolbox_api.Node toShareNode = await getNode(nodePath);
            String uuid = await cloud.generateUUID();
            Event toPostEvent = Event(
                id_event: uuid,
                tlp: toShareNode.visibility.toValueString().toUpperCase());
            toPostEvent.setOwner = senderUserId;
            toPostEvent.setType = 'user';
            toPostEvent.encoding = 'ascii';
            //EncryptNode
            try {
              Map<dynamic, dynamic> jsonify =
                  await encryptCloudData(toShareNode);
              toPostEvent.setContent = jsonEncode(jsonify);
            } catch (e) {
              log("ISSUE ENCRYPTING DATA");
              toPostEvent.setContent =
                  json.encode(await convertNodeToJsonString(toShareNode));
            }
            await cloud.createEvent(senderUserId, toPostEvent);
          } else {
            return false;
            //throw CloudException('There is no active cloud agreement');
          }
        } catch (e) {
          return false;
          //throw CloudException('There is no active cloud agreement');
        }
        return true;
      } else {
        return false;
        //return false;
      }
    } catch (e) {
      return false;
      //throw ReplicationException('Key Node not found');
    }
  }

  /// TODO NO NEED FOR THIS getSharedNodes
  @override
  Future<bool> getSharedNodes(
      String receiverUserId, String senderUserId) async {
    log("******************************************************************");
    log("***************CURRENTLY NOT GETTING SHARED NODES*****************");
    /*
    /// get receiverNodes
    /// cloud API retrieves also the shared ones
    /// check the Event owner to differentiate
    try {
      log("GET SHARED NODES BETWEEN TWO PAIRED USER/DEVICE - FUNCTION getSharedNodes");
      toolbox_api.Node node = await getNode(':Local:Pairing:$senderUserId');
      String encryptedKey = await node
          .getValue('key')
          .then((value) => value!.getValue("en").toString());
      String agreeValue = await node
          .getValue('agreement')
          .then((value) => value!.getValue("en").toString());
      if (agreeValue == 'in' || agreeValue == 'both') {
        try {
          List<String> agreements = await cloud.getMergedAccounts(senderUserId);
          if (agreements.contains(receiverUserId) == true) {
            /// check if our user has publicKey
            /// else the key will be clear
            User actualUser = await cloud.getUser(receiverUserId);
            var publicKey = actualUser.getPublicKey;
            if (publicKey != null) {
              publicKey = publicKey.toString();
              try {
                final keyVal = Enc.Key.fromUtf8(publicKey);
                final enc =
                    Enc.Encrypter(Enc.AES(keyVal, mode: Enc.AESMode.cbc));
                final iv = Enc.IV.fromLength(16);
                //DECRYPT DATA
                encryptedKey = enc.decrypt(
                    encryptedKey.split(':')[1] as Enc.Encrypted,
                    iv: iv);
              } catch (e) {
                log("ISSUE ENCRYPTING DATA - FUNCTION getSharedNodes");
                encryptedKey = publicKey;
              }
            }

            /// get user nodes
            List<String> allEvents = await cloud.getUserEvents(receiverUserId);
            List<Event> waitingEvents = [];

            log("GET USER SHARED EVENTS - FUNCTION getSharedNodes");
            for (var event in allEvents) {
              log("STARTING SHARED EVENTS - FUNCTION getSharedNodes");
              Event newEvent =
                  await cloud.getSingleUserEvent(receiverUserId, event);
              var owner = newEvent.getOwner;
              String id = newEvent.id_event;
              var type = newEvent.type;
              if (owner != receiverUserId) {
                if (newEvent.content != null) {
                  if (type!.toLowerCase() == "keyvalue") {
                    log('CLOUD TO LOCAL REPLICATION - NO E2EE - FUNCTION getSharedNodes');
                    toolbox_api.Node node =
                        convertJsonStringToNode(newEvent.content!);
                    await checkParents(node.path);
                    toolbox_api.Node pairedNode =
                        toolbox_api.NodeImpl(node.path, _devicesPairing);
                    pairedNode.visibility = node.visibility;
                    Map<String, toolbox_api.NodeValue> nodeData =
                        await node.getValues();
                    for (var entry in nodeData.entries) {
                      pairedNode.addOrUpdateValue(entry.value);
                    }
                    log("[PAIRING 1] NODE TO BE UPDATED - FUNCTION getSharedNodes");
                    log(pairedNode);
                    await updateLocalNodeWithCloudNode(pairedNode, _devicesPairing);
                  }
                  if (type.toLowerCase() == "user_object") {
                    log('CLOUD TO LOCAL REPLICATION - E2EE - FUNCTION getSharedNodes');
                    if(_enableEncryption){
                      try {
                      ///CREATE A SEARCH CRITERIA TO FIND KEY WITH GIVEN PATH
                      ///ELSE CREATE A NODE WITH VISIBILITY AMBER
                      ///AND REPLICATE
                      Map<dynamic, dynamic> jsonify =
                          await decryptCloudData(newEvent);
                      newEvent.setContent = jsonEncode(jsonify);
                      toolbox_api.Node node =
                          convertJsonStringToNode(newEvent.content!);
                      await checkParents(node.path);
                      toolbox_api.Node pairedNode =
                          toolbox_api.NodeImpl(node.path, _devicesPairing);
                      pairedNode.visibility = node.visibility;
                      Map<String, toolbox_api.NodeValue> nodeData =
                          await node.getValues();
                      for (var entry in nodeData.entries) {
                        pairedNode.addOrUpdateValue(entry.value);
                      }
                      log("[PAIRING 2] NODE TO BE UPDATED - FUNCTION getSharedNodes");
                      log(pairedNode);
                      await updateLocalNodeWithCloudNode(
                          pairedNode, _devicesPairing);
                      } catch (e) {
                        /// KAY MAY NOT BE REPLICATED YET - CREATE NEW LIST AND DO SAME OPERATION
                        waitingEvents.add(newEvent);
                        log('FAILURE GETTING KEYS FOR NODE - FUNCTION getSharedNodes: $id');
                        log(newEvent);
                      }
                    } else {
                      try {
                        toolbox_api.Node node =
                        convertJsonStringToNode(newEvent.content!);
                        await checkParents(node.path);
                        toolbox_api.Node pairedNode =
                        toolbox_api.NodeImpl(node.path, _devicesPairing);
                        pairedNode.visibility = node.visibility;
                        Map<String, toolbox_api.NodeValue> nodeData =
                        await node.getValues();
                        for (var entry in nodeData.entries) {
                          pairedNode.addOrUpdateValue(entry.value);
                        }
                        log("[PAIRING 2] NODE TO BE UPDATED - FUNCTION getSharedNodes");
                        log(pairedNode);
                        await updateLocalNodeWithCloudNode( pairedNode, _devicesPairing);
                      } catch (e) {
                        /// KAY MAY NOT BE REPLICATED YET - CREATE NEW LIST AND DO SAME OPERATION
                        waitingEvents.add(newEvent);
                        log('ERROR WITHIN FUNCTION getSharedNodes: $id :'+ e.toString());
                        log(newEvent);
                      }
                    }
                  }
                } else {
                  log("THE OWNER OF THE SHARED NODE IS - FUNCTION getSharedNodes: " + owner!);
                }
              }
              for (var waiting in waitingEvents) {
                log("STARTING WAITING EVENTS - FUNCTION getSharedNodes");
                String waitingId = waiting.id_event.toString();
                var owner = waiting.getOwner;
                if (owner != receiverUserId) {
                  log("WAITING EVENT - FUNCTION getSharedNodes: $waitingId");
                  if(_enableEncryption) {
                    try {
                      ///CREATE A SEARCH CRITERIA TO FIND KEY WITH GIVEN PATH
                      ///ELSE CREATE A NODE WITH VISIBILITY AMBER
                      ///AND REPLICATE
                      Map<dynamic, dynamic> jsonify = await decryptCloudData(
                        waiting);
                      waiting.setContent = jsonEncode(jsonify);
                      toolbox_api.Node node =
                        convertJsonStringToNode(waiting.content!);
                      await checkParents(node.path);
                      toolbox_api.Node pairedNode =
                      toolbox_api.NodeImpl(node.path, _devicesPairing);
                      pairedNode.visibility = node.visibility;
                      Map<String, toolbox_api.NodeValue> nodeData =
                      await node.getValues();
                      for (var entry in nodeData.entries) {
                        pairedNode.addOrUpdateValue(entry.value);
                      }
                      log(
                        "[PAIRING 3] NODE TO BE UPDATED - FUNCTION getSharedNodes");
                      log(pairedNode);
                      await updateLocalNodeWithCloudNode(
                        pairedNode, _devicesPairing);
                    } catch (e) {
                      log(
                        'WAITING EVENTS - FAILURE GETTING KEYS FOR NODE - FUNCTION getSharedNodes: $waitingId');
                    }
                  } else {
                    try {
                      toolbox_api.Node node =
                        convertJsonStringToNode(waiting.content!);
                      await checkParents(node.path);
                      toolbox_api.Node pairedNode =
                      toolbox_api.NodeImpl(node.path, _devicesPairing);
                      pairedNode.visibility = node.visibility;
                      Map<String, toolbox_api.NodeValue> nodeData =
                      await node.getValues();
                      for (var entry in nodeData.entries) {
                        pairedNode.addOrUpdateValue(entry.value);
                      }
                      log("[PAIRING 4] NODE TO BE UPDATED - FUNCTION getSharedNodes");
                      log(pairedNode);
                      await updateLocalNodeWithCloudNode(pairedNode, _devicesPairing);
                    } catch (e) {
                      log("GET SHARED NODES FAILURE - FUNCTION getSharedNodes");
                      log(e.toString());
                    }
                  }
                }
              }
            }
          }
        } catch (e) {
          log(e.toString());
          return false;
        }
      }
    } catch (e) {
      log(e.toString());
      return false;
    }*/
    return true;
  }

  @override
  Future<bool> createCloudUser(String userId,
      [String? email,
      String? access,
      String? expires,
      String? name,
      String? publicKey]) async {
    try {
      await cloud.createUser(
          userId.toString(), email, access, expires, name, publicKey);
      return true;
    } catch (e) {
      return false;
    }
  }

  /// UTILS
  /// IN THIS SECTION ARE REPETITIVE TASKS
  /// AVOID DUPLICATION CODE

  /* INIT */
  @override
  Future<void> initGeigerStorage() async {
    log("INIT GEIGER STORAGE");
    try {
      GeigerApi api = await _initGeigerApi();
      log(api.toString());
      _storageController = api.storage;
      log(_storageController.toString());
    } catch (e) {
      log("DATABASE CONNECTION ERROR FROM LOCALSTORAGE: " + e.toString());
      rethrow;
    }
    log("END INIT GEIGER STORAGE");
    //log('[REPLICATION] INIT GEIGER STORAGE');
    //WidgetsFlutterBinding.ensureInitialized();

    //await toolbox_api.StorageMapper.initDatabaseExpander();
    //_storageController = toolbox_api.GenericController('Cloud-Replication', toolbox_api.SqliteMapper('dbFileName.bd'));
    //return _storageController = toolbox_api.GenericController(
    //'Cloud-Replication', toolbox_api.DummyMapper('Cloud-Replication'));
    //return _storageController = storage;
  }

  Future<GeigerApi> _initGeigerApi() async {
    log("INIT GEIGER API");
    try {
      // LN: Due to the bug in the GeigerAPI, the internal plugin cannot see the Storage Event
      // We will try to initialize replication as an external plugin
      // When the bug has been fixed, we should reverse back to be an internal plugin
      _localMaster = (await getGeigerApi(
          _replicationAPI, GeigerApi.masterId, Declaration.doNotShareData))!;
      // _localMaster = (await getGeigerApi(
      //     "cloud_adapter", 'atos_replication_id', Declaration.doShareData))!;
      return _localMaster;
    } catch (e, s) {
      log("ERROR FROM GEIGERAPI: $e, Stack: $s");
      throw toolbox_api.StorageException("ERROR");
    }
  }

  // Register the External Plugin listener to listen all messages (events)
  // ignore: unused_element
  Future<bool> _registerListener() async {
    log("Plugin listener called");
    String id = _localMaster.id;
    if (_isPluginListenerRegistered == true) {
      log(
          'External Plugin ${_messageListener.hashCode} has been registered already!');
      return true;
    } else {
      if (_messageListener == null) {
        _messageListener = ReplicationMessageListener(id);
        log('External Plugin ${_messageListener.hashCode}');
      }
      try {
        // await _pluginAPI!
        //     .registerListener(handledEvents, pluginListener!); // This should be correct one
        _localMaster
            .registerListener([MessageType.allEvents], _messageListener!);
        _localMaster
            .registerListener([MessageType.registerPlugin], _messageListener!);
        log(
            'External Plugin ${_messageListener!.hashCode} has been registered and activated');
        _isPluginListenerRegistered = true;
        return true;
      } catch (e) {
        log('Failed to register listener' + e.toString());
        return false;
      }
    }
  }

  // Dynamically define the handler for each message type
  void addMessageHandler(MessageType type, Function handler) {
    log("addMessageHandler called");
    String id = _localMaster.id;
    if (_messageListener == null) {
      _messageListener = ReplicationMessageListener(id);
      log('External PluginListener: ${_messageListener.hashCode}');
    }
    //don't have an idea all this line of code
    _handledEvents.add(type);
    //pass type and handler to addMessageHandler
    _messageListener!.addMessageHandler(type, handler);
  }

  // Send a simple message which contain only the message type to specific external plugin
  Future<bool> sendAMessageType(
      {required MessageType messageType,
      required String targetId,
      String? id,
      GeigerUrl? pluginUrl}) async {
    try {
      log('Trying to send a message type $messageType');
      Message? request;
      if (pluginUrl != null) {
        //plugin url
        request = Message(
          GeigerApi.masterId,
          targetId,
          messageType,
          pluginUrl,
        );
      } else {
        //toolbox ui url
        final GeigerUrl toolboxUrl =
            GeigerUrl.fromSpec('geiger://toolbox_ui/${id ?? ""}');
        request = Message(
            GeigerApi.masterId, targetId, messageType, toolboxUrl, null, id);
      }

      await _localMaster.sendMessage(request);
      log('A message type $messageType has been sent successfully');
      return true;
    } catch (e) {
      log('Failed to send a message type $messageType' + e.toString());
      return false;
    }
  }

  /*
  * GET ANY NODE -> BASED ON THE PATH
  */
  Future<toolbox_api.Node> getNode(String path) async {
    ///log("GET TOOLBOX STORAGE NODE");
    toolbox_api.Node node = await _storageController.get(path);
    return node;
  }

  // Tells if a node exists in the local storage
  Future<bool> nodeExists(String path) async {
    try {
      toolbox_api.SearchCriteria sensorCriteria =
          toolbox_api.SearchCriteria(searchPath: path);
      List<toolbox_api.Node> scProf =
          await _storageController.search(sensorCriteria);
      if (scProf.isNotEmpty) {
        return true;
      } else {
        return false;
      }
    } catch (e) {
      return false;
    }
  }

  // After replication ends, update with the compared time
  Future<void> updateReplicationNode(DateTime _actual) async {
    /// Should be defined a historic with all replication timestamps?
    /// Add type of replication into the node (full, partial)?
    try {
      //log('UPDATE REPLICATION NODE');
      toolbox_api.Node updateRepNode =
          await getNode(':Local:Replication:LastReplication');
      updateRepNode.lastModified = _actual.millisecondsSinceEpoch;
      await _storageController.update(updateRepNode);
    } catch (e) {
      //log('CREATE NEW REPLICATION NODE');
      toolbox_api.Node parentNode =
          toolbox_api.NodeImpl(':Local:Replication', _replicationAPI);
      await _storageController.addOrUpdate(parentNode);
      toolbox_api.Node newRepNode = toolbox_api.NodeImpl(
          ':Local:Replication:LastReplication', _replicationAPI);

      newRepNode.lastModified = _actual.millisecondsSinceEpoch;
      await _storageController.addOrUpdate(newRepNode);
    }
    log('[REPLICATION NODE] UPDATED');
  }

  Future<void> updateLocalNodeWithCloudNode(
      toolbox_api.Node eventNode, String username) async {
    log('UPDATE LOCAL NODE WITH CLOUD NODE');
    String _nodePath = eventNode.path;
    try {
      toolbox_api.Node inLocal = await getNode(_nodePath);
      //CHECK TIMESTAMP
      DateTime cloud =
          DateTime.fromMillisecondsSinceEpoch(eventNode.lastModified);
      DateTime local =
          DateTime.fromMillisecondsSinceEpoch(inLocal.lastModified);
      Duration _diff = local.difference(cloud);
      if (_diff.inMilliseconds <= 0 &&
          inLocal.owner != GeigerApi.masterId.toString()) {
        if (inLocal.path.toLowerCase().startsWith(":keys") == true) {
          log("KEY VALUE");

          /// KEYS WORK DIFFERENT WITH A SEARCH CRITERIA
          /// COMPARE PATHS && KEYS
          String _localKeyPath = '';
          try {
            _localKeyPath = await inLocal
                .getValue('path')
                .then((value) => value!.getValue("en").toString());
          } catch (e) {
            _localKeyPath = '';
          }
          String _cloudKeyPath = '';
          try {
            _cloudKeyPath = await eventNode
                .getValue('path')
                .then((value) => value!.getValue("en").toString());
          } catch (e) {
            _cloudKeyPath = '';
          }
          if (_cloudKeyPath.isNotEmpty &&
              _localKeyPath.isNotEmpty &&
              _cloudKeyPath != _localKeyPath) {
            log("KEY VALUE DO NOT MATCH");

            /// VALUE IS NOT THE SAME
            /// CREATE A NEW UUID
            String newUUID = Uuid().v4();
            String fullPath = eventNode.parentPath! + ':' + newUUID;
            toolbox_api.Node renamed =
                toolbox_api.NodeImpl(fullPath, eventNode.owner);
            toolbox_api.Visibility? checkerVisible =
                toolbox_api.VisibilityExtension.valueOf("amber");
            if (checkerVisible != null) {
              renamed.visibility = checkerVisible;
            }
            Map<String, toolbox_api.NodeValue> newValues =
                await eventNode.getValues();
            for (var entry in newValues.entries) {
              renamed.addOrUpdateValue(entry.value);
            }
            await _storageController.addOrUpdate(renamed);
          } else {
            log("PATH DO NOT MATCH");
            log(eventNode.toString());
            await _storageController.addOrUpdate(eventNode);
          }
        } else {
          /// IF LOCAL EVENT HAS CHILDREN. GET AND UPDATE
          Map<String, toolbox_api.Node> children = await inLocal.getChildren();
          for (var child in children.entries) {
            eventNode.addChild(child.value);
          }
          await _storageController.addOrUpdate(eventNode);
        }
      } else {
        log(GeigerApi.masterId.toString());
      }
    } catch (e) {
      log('NODE NOT FOUND - CREATE ONE');
      eventNode.owner = username;
      await _storageController.add(eventNode);
    }
  }

  Future<void> updateLocalNodeWithNewNode(
      String path, toolbox_api.Node node, DateTime cloud) async {
    //log('CHECK NODES');
    try {
      toolbox_api.Node inLocal = await getNode(path);
      DateTime local =
          DateTime.fromMillisecondsSinceEpoch(inLocal.lastModified);
      Duration _diff = local.difference(cloud);
      if (_diff.inMilliseconds <= 0) {
        await _storageController.update(node);
      }
    } catch (e) {
      //log('NODE DO NOT EXIST - CREATE ONE');
      await _storageController.add(node);
    }
  }

  Future<void> updateTLPWhiteEvents(bool _fullRep, DateTime _fromDate) async {
    log("UPDATE TLP WHITE EVENTS");
    List<Event> freeEvents;
    // FILTER BY DATE
    if (_fullRep == true) {
      freeEvents = await cloud.getTLPWhiteEvents();
    } else {
      var filter =
          DateFormat("yyyy-MM-dd'T'hh:mm:ss.SSSS'Z'").format(_fromDate);
      freeEvents = await cloud.getTLPWhiteEventsDateFilter(filter.toString());
    }
    for (var free in freeEvents) {
      /// UUID WILL BE THE CLOUD ONE
      /// DATA MAY NOT COME FROM STORAGE AND USER ANOTHER UUID
      //String typeTLP = free.type.toString();
      String uuid = free.id_event;
      DateTime cloudTimestamp = DateTime.parse(free.last_modified.toString());
      if (free.content.toString().isNotEmpty) {
        try {
          ///var content = jsonDecode(free.content!);
          String nodePath = ':Global:$uuid';

          /// CHECK IF typeTLP node created
          try {
            // ignore: unused_local_variable
            toolbox_api.Node typeChecker = await getNode(':Global');
          } catch (e) {
            /// CREATE NODE FOR TYPE of TLP WHITE EVENT
            toolbox_api.Node typeChecker =
                toolbox_api.NodeImpl(':Global', free.owner.toString());
            toolbox_api.Visibility? checkerVisible =
                toolbox_api.VisibilityExtension.valueOf("white");
            if (checkerVisible != null) {
              typeChecker.visibility = checkerVisible;
            }
            await _storageController.add(typeChecker);
          }
          toolbox_api.Node newRepNode =
              toolbox_api.NodeImpl(nodePath, free.owner.toString());
          toolbox_api.Visibility? visible =
              toolbox_api.VisibilityExtension.valueOf(free.getTlp.toString());
          if (visible != null) {
            newRepNode.visibility = visible;
          }
          //LOOP ALL ELEMENTS
          //newRepNode.addOrUpdateValue(
          // toolbox_api.NodeValueImpl("tlpWHITE", free.content!));
          //await _storageController.update(newRepNode);
          Map<dynamic, dynamic> mapper = json.decode(free.content!);
          mapper.forEach((key, value) {
            newRepNode.addOrUpdateValue(
                toolbox_api.NodeValueImpl(key, value.toString()));
          });
          await updateLocalNodeWithNewNode(
              nodePath, newRepNode, cloudTimestamp);
        } catch (e) {
          //log("CONTENT TYPE NOT IN JSON FORMAT");
        }
      }
    }
  }

  @override
  Future<void> updateThreats() async {
    log('[UPDATE  GEIGER THREATS]');
    log("CHECK PARENT NODE: :GLOBAL:THREATS");
    toolbox_api.Node typeChecker;
    try {
      typeChecker = await getNode(':Global:threats');
    } catch (e) {
      /// CREATE NODE FOR TYPE of TLP WHITE EVENT
      typeChecker = toolbox_api.NodeImpl(':Global:threats', _replicationAPI);
      toolbox_api.Visibility? checkerVisible =
          toolbox_api.VisibilityExtension.valueOf("white");
      if (checkerVisible != null) {
        typeChecker.visibility = checkerVisible;
      }
      await _storageController.add(typeChecker);
    }
    List<Threats> threats = await cloud.getThreats();
    for (var threat in threats) {
      String? uuid = threat.getUuid;
      String? name = threat.getName;
      String? value = threat.getThreat;
      try {
        toolbox_api.Node compare = await getNode(':Global:threats:$uuid');
        compare.addOrUpdateValue(toolbox_api.NodeValueImpl('name', name!));
        compare.addOrUpdateValue(
            toolbox_api.NodeValueImpl('GEIGER_threat', value!));
        await _storageController.addOrUpdate(compare);
      } catch (e) {
        toolbox_api.Node compare =
            toolbox_api.NodeImpl(':Global:threats:$uuid', _replicationAPI);
        compare.addOrUpdateValue(toolbox_api.NodeValueImpl('name', name!));
        compare.addOrUpdateValue(
            toolbox_api.NodeValueImpl('GEIGER_threat', value!));
        await _storageController.addOrUpdate(compare);
      }
    }
  }

  /// STORE :Global:threats
  @override
  Future<void> updateThreatWeights() async {
    log('[UPDATE THREAT WEIGHTS]');
    log("CHECK PARENT NODE: :GLOBAL:PROFILES");
    try {
      // ignore: unused_local_variable
      toolbox_api.Node profileChecker = await getNode(':Global:profiles');
    } catch (e) {
      /// CREATE NODE FOR TYPE of TLP WHITE EVENT
      toolbox_api.Node profileChecker =
          toolbox_api.NodeImpl(':Global:profiles', _replicationAPI);
      toolbox_api.Visibility? checkerVisible =
          toolbox_api.VisibilityExtension.valueOf("white");
      if (checkerVisible != null) {
        profileChecker.visibility = checkerVisible;
      }
      await _storageController.add(profileChecker);
    }

    log(
        "CHECK PARENT NODES SUB NODE - DIGITALLY DEPENDENT, DIGITALLY BASED, DIGITAL ENABLER");
    List<String> profiles = [
      "digitally dependent",
      "digitally based",
      "digitally enabler"
    ];
    List<toolbox_api.Node> profileNodes = [];
    for (var profile in profiles) {
      /// search criteria for each profile
      toolbox_api.SearchCriteria profileCriteria = toolbox_api.SearchCriteria(
          searchPath: ':Global:profiles:', key: 'name', value: profile);
      List<toolbox_api.Node> scProf =
          await _storageController.search(profileCriteria);
      if (scProf.isNotEmpty) {
        profileNodes.add(scProf[0]);
      } else {
        /// CREATE NEW
        String newUUID = Uuid().v4();
        toolbox_api.Node newProfile =
            toolbox_api.NodeImpl(':Global:profiles:$newUUID', _replicationAPI);
        newProfile.addOrUpdateValue(toolbox_api.NodeValueImpl('name', profile));
        await _storageController.add(newProfile);
        profileNodes.add(newProfile);
      }
    }
    List<ThreatWeights> weights = await cloud.getThreatWeights();
    for (var weight in weights) {
      ThreatDict? data = weight.threatDict;
      if (data != null) {
        Map<String, dynamic> mapper = data.toJson();
        for (var d in mapper.entries) {
          String key = d.key;
          List<double> value = d.value;
          //mapper.forEach((key, value) async {
          toolbox_api.SearchCriteria criteria = toolbox_api.SearchCriteria(
              searchPath: ':Global:threats:',
              key: 'GEIGER_threat',
              value: '${key[0].toUpperCase()}${key.substring(1)}');
          List<toolbox_api.Node> scNode =
              await _storageController.search(criteria);
          List<double> val = value;
          String path;
          if (scNode.isNotEmpty) {
            for (var element in scNode) {
              toolbox_api.Node newThreat = element;
              path = newThreat.name;
              toolbox_api.NodeValue? nameVal =
                  (await newThreat.getValue('name'));
              toolbox_api.NodeValue? threatVal =
                  (await newThreat.getValue('GEIGER_threat'));
              if (nameVal!.value == threatVal!.value) {
                int i = 0;
                for (var v in val) {
                  profileNodes[i].addOrUpdateValue(
                      toolbox_api.NodeValueImpl(path, v.toString()));
                  i++;
                }
              }
            }
          }
        }
        //});
      }
    }

    /// store updated nodes in toolbox
    log("STORE/UPDATE PROFILE NODES");
    for (var node in profileNodes) {
      await _storageController.addOrUpdate(node);
    }
  }

  @override
  Future<void> updateRecommendations() async {
    log('[UPDATE RECOMMENDATIONS]');
    toolbox_api.Node typeChecker;
    try {
      typeChecker = await getNode(':Global:Recommendations');
    } catch (e) {
      /// CREATE NODE FOR TYPE of TLP WHITE EVENT
      typeChecker =
          toolbox_api.NodeImpl(':Global:Recommendations', _replicationAPI);
      toolbox_api.Visibility? checkerVisible =
          toolbox_api.VisibilityExtension.valueOf("white");
      if (checkerVisible != null) {
        typeChecker.visibility = checkerVisible;
      }
      await _storageController.add(typeChecker);
    }
    List<Recommendation> recommendations = await cloud.getRecommendations();
    for (var rec in recommendations) {
      String? id = rec.getId;
      if (id != null) {
        try {
          toolbox_api.Node checker =
              await getNode(':Global:Recommendations:$id');
          Map<String, dynamic> mapper = rec.toJson();
          for (var r in mapper.entries) {
            if (r.key != "relatedThreatsWeights" && r.key != "Steps") {
              checker.addOrUpdateValue(
                  toolbox_api.NodeValueImpl(r.key, r.value.toString()));
            } else {
              if (r.key == "Steps") {
                String toAdd = "";
                List<String> steps = r.value;
                for (var step in steps) {
                  if (toAdd.isEmpty) {
                    toAdd = toAdd + "[" + '"' + step + '"';
                  } else {
                    toAdd = toAdd + "," + '"' + step + '"';
                  }
                }
                toAdd = toAdd + "]";
                checker
                    .addOrUpdateValue(toolbox_api.NodeValueImpl(r.key, toAdd));
              }
              if (r.key == "relatedThreatsWeights") {
                /// IF RELATED THREATS WEIGHTS
                /// SEARCH CRITERIA TO GET NODE ID FOR KEY VALUE
                String data = "";
                List<String> related = r.value;
                List<String> splitted = related[0].split(",");
                for (var split in splitted) {
                  try {
                    List<String> splitThreat = split.split(":");
                    String threat = splitThreat[0].trim();
                    String threatValue = splitThreat[1].trim();
                    toolbox_api.SearchCriteria profileCriteria =
                        toolbox_api.SearchCriteria(
                            searchPath: ':Global:threats:',
                            key: 'GEIGER_threat',
                            value: threat);
                    List<toolbox_api.Node> scProf =
                        await _storageController.search(profileCriteria);
                    if (scProf.isNotEmpty) {
                      for (var element in scProf) {
                        toolbox_api.Node threatNode = element;
                        String uuid = threatNode.name;
                        toolbox_api.NodeValue? nameVal =
                            (await threatNode.getValue('name'));
                        toolbox_api.NodeValue? threatVal =
                            (await threatNode.getValue('GEIGER_threat'));
                        if (nameVal!.value == threatVal!.value) {
                          if (data.isEmpty == true) {
                            data = data + uuid + "," + threatValue;
                          } else {
                            data = data + ";" + uuid + "," + threatValue;
                          }
                        }
                      }
                    }
                  } catch (e) {
                    log("EXCEPTION GETTING RECOMMENDATIONS THREAT WEIGHTS");
                  }
                }
                checker
                    .addOrUpdateValue(toolbox_api.NodeValueImpl(r.key, data));
              }
            }
          }
          await _storageController.addOrUpdate(checker);
        } catch (e) {
          toolbox_api.Node newThreatNode = toolbox_api.NodeImpl(
              ':Global:Recommendations:$id', _replicationAPI);
          toolbox_api.Visibility? visible =
              toolbox_api.VisibilityExtension.valueOf("white");
          if (visible != null) {
            newThreatNode.visibility = visible;
          }
          Map<String, dynamic> mapper = rec.toJson();
          for (var r in mapper.entries) {
            if (r.key != "relatedThreatsWeights" && r.key != "Steps") {
              newThreatNode.addOrUpdateValue(
                  toolbox_api.NodeValueImpl(r.key, r.value.toString()));
            } else {
              if (r.key == "Steps") {
                String toAdd = "";
                List<String> steps = r.value;
                for (var step in steps) {
                  if (toAdd.isEmpty) {
                    toAdd = toAdd + "[" + '"' + step + '"';
                  } else {
                    toAdd = toAdd + "," + '"' + step + '"';
                  }
                }
                toAdd = toAdd + "]";
                newThreatNode
                    .addOrUpdateValue(toolbox_api.NodeValueImpl(r.key, toAdd));
              }
              if (r.key == "relatedThreatsWeights") {
                /// IF RELATED THREATS WEIGHTS
                /// SEARCH CRITERIA TO GET NODE ID FOR KEY VALUE
                String data = "";
                List<String> related = r.value;
                List<String> splitted = related[0].split(";");
                for (var split in splitted) {
                  try {
                    List<String> splitThreat = split.split(",");
                    String threat = splitThreat[0].trim();
                    String threatValue = splitThreat[1].trim();
                    toolbox_api.SearchCriteria profileCriteria =
                        toolbox_api.SearchCriteria(
                            searchPath: ':Global:threats:',
                            key: 'GEIGER_threat',
                            value: threat);
                    List<toolbox_api.Node> scProf =
                        await _storageController.search(profileCriteria);
                    if (scProf.isNotEmpty) {
                      for (var element in scProf) {
                        toolbox_api.Node threatNode = element;
                        String uuid = threatNode.name;
                        toolbox_api.NodeValue? nameVal =
                            (await threatNode.getValue('name'));
                        toolbox_api.NodeValue? threatVal =
                            (await threatNode.getValue('GEIGER_threat'));
                        if (nameVal!.value == threatVal!.value) {
                          if (data.isEmpty == true) {
                            data = data + uuid + "," + threatValue;
                          } else {
                            data = data + ";" + uuid + "," + threatValue;
                          }
                        }
                      }
                    }
                  } catch (e) {
                    log("EXCEPTION GETTING RECOMMENDATIONS THREAT WEIGHTS");
                  }
                }
                newThreatNode
                    .addOrUpdateValue(toolbox_api.NodeValueImpl(r.key, data));
              }
            }
          }
          await _storageController.addOrUpdate(newThreatNode);
        }
      }
    }
  }

  Future<List<toolbox_api.Node>> getAllNodes(bool cleanup) async {
    log("GET ALL NODES IN A RECURSIVE WAY " +
        (cleanup ? 'TO CLEAN' : 'TO ADD'));
    List<toolbox_api.Node> nodeList = [];
    try {
      toolbox_api.Node root1 = await getNode(':Devices');
      await getRecursiveNodes(root1, nodeList, cleanup);
      toolbox_api.Node root2 = await getNode(':Users');
      await getRecursiveNodes(root2, nodeList, cleanup);
      toolbox_api.Node root3 = await getNode(':Keys');
      await getRecursiveNodes(root3, nodeList, cleanup);
      toolbox_api.Node root4 = await getNode(':Enterprise');
      await getRecursiveNodes(root4, nodeList, cleanup);
      toolbox_api.Node root5 = await getNode(':Chatbot');
      await getRecursiveNodes(root5, nodeList, cleanup);
    } catch (e) {
      log(e.toString());
    }
    log("NODE COUNT IS " + nodeList.length.toString());
    return nodeList;
  }

  Future<List<toolbox_api.Node>> getRecursiveNodes(
      toolbox_api.Node node, List<toolbox_api.Node> list, bool cleanup) async {
    String dataPath = node.path.toString();
    if (dataPath != ":" &&
        dataPath != ":Users" &&
        dataPath != ":Devices" &&
        dataPath != ":Enterprise" &&
        dataPath != ":Chatbot" &&
        dataPath != ":Keys" &&
        dataPath != ":Global" &&
        dataPath != ":Local") {
      log("NODE THAT CAN BE " + (cleanup ? 'DELETED' : 'ADDED'));
      log(dataPath);
      list.add(node);
    }
    Map<String, toolbox_api.Node> children = await node.getChildren();
    if (children.isNotEmpty) {
      for (var value in children.values) {
        list = await getRecursiveNodes(value, list, cleanup);
      }
    }
    return list;
  }

  Future<List<toolbox_api.Node>> getAllNodesLastModified(
      DateTime filter) async {
    String lastModified = filter.millisecondsSinceEpoch.toString();
    toolbox_api.SearchCriteria criteria =
        toolbox_api.SearchCriteria(nodeValueLastModified: lastModified);
    List<toolbox_api.Node> nodeList = [];
    List<toolbox_api.Node> preNodeList =
        await _storageController.search(criteria);
    for (var node in preNodeList) {
      String dataPath = node.path.toString();
      if (dataPath != ":" &&
          dataPath != ":Users" &&
          dataPath != ":Devices" &&
          dataPath != ":Enterprise" &&
          dataPath != ":Keys" &&
          dataPath != ":Global" &&
          dataPath != ":Local") {
        log("NODE THAT CAN BE ADDED");
        nodeList.add(node);
      }
    }
    return nodeList;
  }

  Future<void> cleanData(DateTime actual) async {
    log('TIMING STRATEGY - REMOVE DATA - 180 DAYS LIMIT');

    /// Check if first replication is 180 days older

    /// COMPLETE
    /// RECURSIVE WAY
    /// CHECK TIMESTAMP > 180 DAYS: DELETE
    List<toolbox_api.Node> checkingData = await getAllNodes(true);

    /// SORT NODES BY TIMESTAMP
    checkingData.sort((a, b) =>
        a.lastModified.toString().compareTo(b.lastModified.toString()));
    for (var node in checkingData) {
      DateTime nodeTime =
          DateTime.fromMillisecondsSinceEpoch(node.lastModified);
      Duration checker = actual.difference(nodeTime);
      if (checker.inDays > 180) {
        //REPLICATION TIMING STRATEGIES - DELETE NODE
        String path = node.path.toString();
        await _storageController.delete(path);
        log('NODE WITH PATH $path HAS BEEN REMOVED');
      }
    }
  }

  @override
  Future<bool> checkConsent(toolbox_api.Node node, String username) async {
    bool consent = false;

    /// BY DEFAULT
    /// GET CONSENT NODE
    /// IF NOT - ASK AND CREATE
    try {
      //IF GET NODE - USER HAS NOT AGREED FOREVER - CONSENT FALSE
      toolbox_api.Node consentNode = await getNode(':Local:Consent:$username');
      String consentValue = await consentNode
          .getValue('consent')
          .then((value) => value!.getValue("en").toString());
      if (consentValue == 'no_forever') {
        consent = false;
      } else if (consentValue == 'no_once' || consentValue == "once") {}
    } catch (e) {
      log("CONSENT NODE NOT FOUND. CREATE CONSENT PARENT NODE");
      toolbox_api.Node consentNode =
          toolbox_api.NodeImpl(':Local:Consent', _replicationAPI);
      await _storageController.add(consentNode);

      /// SEND MESSAGE
      /*GeigerUrl? url;
      try {
        url = GeigerUrl(null, GeigerApi.masterId, 'geiger://plugin/path');
      } catch (e) {
        log(e.toString());
      }
      MessageType messageType = MessageType.storageEvent;
      Message message = Message('ReplicationController', 'uiId', messageType, url);
      await _localMaster.sendMessage(message);*/
      toolbox_api.Node userConsent =
          toolbox_api.NodeImpl(':Local:Consent:$username', _replicationAPI);
      userConsent.addOrUpdateValue(
          toolbox_api.NodeValueImpl("consent", 'consentData'));
      if ('consentData' == 'no_forever') {
        consent = false;
      } else if ('consentData' == 'no_once' || 'consentData' == 'once') {
        toolbox_api.Visibility? visible =
            toolbox_api.VisibilityExtension.valueOf("red");
        if (visible != null) {
          userConsent.visibility = visible;
        }
        if ('consentData' == 'once') {
          consent = true;
        } else {
          consent = false;
        }
      } else {
        log("CONSENT VALUE DOES NOT MATCH");

        /// throw ReplicationException("NOT MATCH CONSENT VALUE");
      }
      await _storageController.add(userConsent);
    }
    return consent;

    /*String tlp = node.visibility.toValueString();
    if (tlp.toLowerCase() == 'white') {
      consent = true;
    } else {
      //CHECK CONSENT
      try {
        //IF GET NODE - USER HAS NOT AGREED FOREVER - CONSENT FALSE
        toolbox_api.Node consentNode = await getNode(':Local:Consent:$username');
        String consentValue = await consentNode
          .getValue('consent')
          .then((value) => value!.getValue("en").toString());
        if (consentValue=='no_forever') {
          return false;
        } else if(consentValue == 'no_once' || consentValue == "once") {
          /// If doen't agree once
          /// ASK AGAIN
          String consentData = "";
          consentNode.addOrUpdateValue(toolbox_api.NodeValueImpl("consent", consentData));
          await _storageController.addOrUpdate(consentNode);
          if (consentValue == 'no_forever') {
            return false;
          } else if(consentValue == 'no_once' || consentValue == 'once') {
            toolbox_api.Visibility? visible =
              toolbox_api.VisibilityExtension.valueOf("red");
            if (visible != null) {
              node.visibility = visible;
            }
            await _storageController.update(node);
            if (consentValue == 'once') {
              return true;
            } else {
              return false;
            }
          } else {
            throw ReplicationException("NOT MATCH CONSENT VALUE");
          }
        } else {
          throw ReplicationException("NOT MATCH CONSENT VALUE");
        }
      } catch (e) {
        log("CONSENT NODE NOT FOUND");
        /// CREATE PARENT CONSENT NODE
        toolbox_api.Node consentNode = toolbox_api.NodeImpl(':Local:Consent','ReplicationController');
        await _storageController.addOrUpdate(consentNode);
        String consentData = "";
        toolbox_api.Node userConsent = toolbox_api.NodeImpl(':Local:Consent:$username','ReplicationController');
        userConsent.addOrUpdateValue(toolbox_api.NodeValueImpl("consent", consentData));
        await _storageController.addOrUpdate(userConsent);
        if (consentData == 'no_forever') {
          return false;
        } else if(consentData == 'no_once' || consentData == 'once') {
          toolbox_api.Visibility? visible =
            toolbox_api.VisibilityExtension.valueOf("red");
          if (visible != null) {
            node.visibility = visible;
          }
          await _storageController.update(node);
          if (consentData == 'once') {
            return true;
          } else {
            return false;
          }
        } else {
          throw ReplicationException("NOT MATCH CONSENT VALUE");
        }
      }
    }
    return consent;*/
  }

  /// TODO NO NEED FOR THIS updatePairedEvent
  /// if a user receives a shared event but with a different owner
  /// checks pairing
  Future<void> updatePairedEvent(String username, Event event) async {
    /*try {
      String owner = event.owner.toString();
      toolbox_api.Node node = await getNode(':Local:Pairing:$owner');
      String encryptedKey = node.getValue('key').toString();
      String agreeValue = node.getValue('agreement').toString();
      if (agreeValue == 'in' || agreeValue == 'both') {
        try {
          List<String> agreements = await cloud.getMergedAccounts(username);
          if (agreements.contains(owner) == true) {
            /// check if our user has publicKey
            /// else the key will be clear
            User actualUser = await cloud.getUser(username);
            var publicKey = actualUser.getPublicKey;
            if (publicKey != null) {
              publicKey = publicKey.toString();

              final keyVal = Enc.Key.fromUtf8(publicKey);
              final enc =
                  Enc.Encrypter(Enc.AES(keyVal, mode: Enc.AESMode.cfb64));
              final iv = Enc.IV.fromLength(16);
              //DECRYPT DATA
              encryptedKey = enc
                  .decrypt(encryptedKey.split(':')[1] as Enc.Encrypted, iv: iv);
            } else {
              encryptedKey = encryptedKey.split(':')[1];
            }

            /// DECRYPT CONTENT AND SET NEW NODE
            final keyVal1 = Enc.Key.fromUtf8(encryptedKey);
            final enc1 =
                Enc.Encrypter(Enc.AES(keyVal1, mode: Enc.AESMode.cfb64));
            final iv1 = Enc.IV.fromLength(16);
            //DECRYPT DATA
            var data = enc1.decrypt(event.content.toString() as Enc.Encrypted,
                iv: iv1);

            /// CREATE NODE
            toolbox_api.Node newSharedNode = convertJsonStringToNode(data);
            toolbox_api.Visibility? visible =
                toolbox_api.VisibilityExtension.valueOf(
                    event.getTlp.toString());
            if (visible != null) {
              newSharedNode.visibility = visible;
            }
            _storageController.update(newSharedNode);
          }
        } catch (e) {
          throw CloudException('There is no active cloud agreement');
        }
      }
    } catch (e) {
      throw toolbox_api.StorageException('No pairing node set up');
    }*/
  }

  toolbox_api.Node convertJsonStringToNode(String content) {
    /// convert string event content to Json
    log("CONVERT JSON TO NODE");
    log(content);
    try {
      Map<String, dynamic> mapper = jsonDecode(content);
      var path = mapper['path'];
      var owner = mapper['owner'];
      var visibility = mapper['visibility'];
      var lastModified = mapper['lastModified'];
      //var name = mapper['name'];
      //var parentPath = mapper['parentPath'];
      //var extendedLastModified = mapper['extendedLastModified'];
      //var tombstone = mapper['tombstone'];
      toolbox_api.Node node = toolbox_api.NodeImpl(path, owner);
      toolbox_api.Visibility? checkerVisible =
          toolbox_api.VisibilityExtension.valueOf(visibility);
      if (checkerVisible != null) {
        node.visibility = checkerVisible;
      }
      node.lastModified = lastModified;
      if (mapper['custom_fields'] != '[]') {
        List customFields = ((mapper['custom_fields']));
        if (customFields.isNotEmpty) {
          for (var custom in customFields) {
            var key = custom['key'];
            var value = custom['value'];
            var nodeDescription = custom['nodeDescription'];
            var valueTranslations = custom['valueTranslations'];
            var descriptionTranslations = custom['descriptionTranslations'];
            toolbox_api.NodeValue nodeValue =
                toolbox_api.NodeValueImpl(key, value.toString());
            for (var translation in valueTranslations) {
              var tr = translation['value'];
              if (tr != null && tr != "") {
                nodeValue.setValue(translation['value'],
                    Locale.fromSubtags(languageCode: translation['key']));
              }
            }
            if (nodeDescription != null) {
              nodeValue.setDescription(nodeDescription);
              for (var description in descriptionTranslations) {
                var dr = description['value'];
                if (dr != null && dr != "") {
                  nodeValue.setValue(description['value'],
                      Locale.fromSubtags(languageCode: description['key']));
                }
              }
            }
            node.addOrUpdateValue(nodeValue);
          }
        }
      }
      return node;
    } catch (e) {
      log(e.toString());
      throw ReplicationException(
          "[REPLICATION EXCEPTION] FAIL PASSING EVENT JSON STRING CONTENT TYPE TO NODE");
    }
  }

  Map<dynamic, dynamic> getPartialJsonString(String content) {
    log("GET PARTIAL DATA");
    try {
      Map<String, dynamic> mapper = jsonDecode(content);
      var path = mapper['path'];
      var owner = mapper['owner'];
      var visibility = mapper['visibility'];
      var lastModified = mapper['lastModified'];
      var name = mapper['name'];
      var parentPath = mapper['parentPath'];
      var extendedLastModified = mapper['extendedLastModified'];
      var tombstone = mapper['tombstone'];
      var converted = {};
      converted['name'] = name;
      converted['parentPath'] = parentPath;
      converted['path'] = path;
      converted['owner'] = owner;
      converted['visibility'] = visibility;
      converted['lastModified'] = lastModified;
      converted['extendedLastModified'] = extendedLastModified;
      converted['tombstone'] = tombstone;
      return converted;
    } catch (e) {
      log(e.toString());
      throw ReplicationException(
          "[REPLICATION EXCEPTION] FAIL GETTING EVENT JSON STRING CONTENT TYPE TO SHORT EVENT");
    }
  }

  Future<Map<dynamic, dynamic>> convertNodeToJsonString(
      toolbox_api.Node node) async {
    log("CONVERT NODE TO JSON");
    try {
      var converted = {};

      /// CONVERT VALUES
      converted['name'] = node.name;
      converted['parentPath'] = node.parentPath;
      converted['path'] = node.path;
      converted['owner'] = node.owner;
      converted['visibility'] = node.visibility.toValueString();
      converted['lastModified'] = node.lastModified;
      //converted['extendedLastModified'] = node.extendedLastModified;
      converted['tombstone'] = node.tombstone;
      converted['custom_fields'] = [];
      Map<String, toolbox_api.NodeValue> mapper = await node.getValues();
      mapper.forEach((key, value) {
        toolbox_api.NodeValue nodeValue = value;
        var loop = {};
        loop['key'] = nodeValue.key;
        loop['value'] = nodeValue.value;
        loop['valueTranslations'] = [];

        Map<Locale, String> valueTranslations = nodeValue.allValueTranslations;
        valueTranslations.forEach((key, value) {
          var valueT = {};
          valueT['key'] = key.toString();
          valueT['value'] = value;
          loop['valueTranslations'].add(valueT);
        });

        loop['nodeDescription'] = nodeValue.getDescription();
        loop['descriptionTranslations'] = [];
        if (nodeValue.getDescription() != null) {
          Map<Locale, String> descriptionTranslations =
              nodeValue.allDescriptionTranslations;
          descriptionTranslations.forEach((key, value) {
            var valueT = {};
            valueT['key'] = key.toString();
            valueT['value'] = value;
            loop['descriptionTranslations'].add(valueT);
          });
        }
        converted['custom_fields'].add(loop);
      });
      //String stringConverted = json.encode(converted);
      return converted;
    } catch (e) {
      log(e.toString());
      throw ReplicationException(
          "[REPLICATION EXCEPTION] FAIL CONVERT NODE TO JSON STRING");
    }
  }

  /// *******************************
  /// UTILS METHOD
  /// *******************************

  /// CHECKS IF THE PARENT NODE IS CREATED
  /// IF NOT CREATES PARENT NODE IN A RECURSIVE WAY
  Future<bool> checkParents(String path) async {
    log("CHECK PARENTS NODE");
    bool exist = false;
    //PATH MUST BE SPLITTED BY ':'
    List<String> splittedPath = path.split(':');
    String createPath = '';
    for (var p in splittedPath) {
      /// GET NODE - IF EXISTS - GO TO NEXT
      /// ELSE - CREATE NODE
      if (p.isNotEmpty) {
        createPath = createPath + ':' + p;
        try {
          // ignore: unused_local_variable
          toolbox_api.Node nodeChecker = await getNode(createPath);
        } catch (e) {
          toolbox_api.Node parentNode =
              toolbox_api.NodeImpl(createPath, _devicesPairing);
          toolbox_api.Visibility? checkerVisible =
              toolbox_api.VisibilityExtension.valueOf("amber");
          if (checkerVisible != null) {
            parentNode.visibility = checkerVisible;
          }
          try {
            await _storageController.addOrUpdate(parentNode);
            log(createPath);
          } catch (e) {
            log(e.toString());
            exist = false;
            return exist;
          }
        }
      }
    }
    exist = true;
    return exist;
  }

  @override
  Future<void> endGeigerStorage() async {
    flushGeigerApiCache();
    await _localMaster.zapState();
    await _localMaster.close();
  }

  @override
  Future<void> updateSecurityDefendersInfo() async {
    //FIRST GET ALL SECURITY DEFENDERS ORGANIZATION
    List<SecurityDefendersOrganization> orgs =
        await cloud.getSecurityDefendersOrganizations();
    for (var org in orgs) {
      //CREATE A SEARCH CRITERIA TO SEE IF COUNTRY DOES EXIST
      String? country = org.location;
      String? id = org.idOrg;
      String? name = org.name;

      if (country != null && id != null) {
        toolbox_api.SearchCriteria criteria = toolbox_api.SearchCriteria(
            searchPath: ':Global:location', value: country);
        List<toolbox_api.Node> nodes =
            await _storageController.search(criteria);
        if (nodes.isEmpty) {
          //NEED TO CREATE COUNTRY
          toolbox_api.Node newCountry = toolbox_api.NodeImpl(
              ':Global:location:' + country, _replicationAPI);
          newCountry
              .addOrUpdateValue(toolbox_api.NodeValueImpl('name', country));
          await _storageController.addOrUpdate(newCountry);
        }
      }
      if (name != null) {
        try {
          // ignore: unused_local_variable
          toolbox_api.Node o =
              await getNode(':Global:securityDefendersOrganizations:' + id!);
        } catch (e) {
          //CREATE NODE
          toolbox_api.Node newOrg = toolbox_api.NodeImpl(
              ':Global:securityDefendersOrganizations:' + id!, _replicationAPI);
          newOrg.addOrUpdateValue(toolbox_api.NodeValueImpl('name', name));
          newOrg.addOrUpdateValue(
              toolbox_api.NodeValueImpl('location', country!));
          await _storageController.addOrUpdate(newOrg);
        }
      }

      //FOR EACH ORG - GET USERS
      List<String> uuids =
          await cloud.getSecurityDefendersOrganizationUsers(id!);
      for (var uuid in uuids) {
        SecurityDefenderUser user =
            await cloud.getSecurityDefenderUser(id, uuid);
        //ADD OR UPDATE
        toolbox_api.Node n = toolbox_api.NodeImpl(
            ':Global:securityDefenders:' + user.id_user!, _replicationAPI);
        n.addOrUpdateValue(toolbox_api.NodeValueImpl('name', user.name!));
        n.addOrUpdateValue(
            toolbox_api.NodeValueImpl('firstname', user.firstname!));
        n.addOrUpdateValue(
            toolbox_api.NodeValueImpl('affiliation', user.affiliation!));
        n.addOrUpdateValue(toolbox_api.NodeValueImpl('phone', user.phone!));
        n.addOrUpdateValue(toolbox_api.NodeValueImpl('email', user.email!));
        n.addOrUpdateValue(toolbox_api.NodeValueImpl('title', user.title!));
        n.addOrUpdateValue(toolbox_api.NodeValueImpl('picture', ''));
        await _storageController.addOrUpdate(n);
      }
    }
  }

  @override
  Future<void> geigerReplicationListener(
      deleteHandler, createHandler, updateHandler, renameHandler) async {
    try {
      userEvents = await cloud.getUserEvents(_username);
      Map<dynamic, dynamic> data;

      for (var cloudEvent in userEvents) {
        try {
          Event newCloudEvent =
              await cloud.getSingleUserEvent(_username, cloudEvent);

          if (newCloudEvent.type != 'keyvalue') {
            if (newCloudEvent.tlp!.toLowerCase() == 'red') {
              try {
                data = await decryptCloudData(newCloudEvent);
              } catch (e) {
                data = getPartialJsonString(newCloudEvent.content!);
                log('FAILURE GETTING KEYS RENAME HANDLER');
              }
            } else {
              data = getPartialJsonString(newCloudEvent.content!);
            }

            if (userEvents.contains(cloudEvent) == false) {
              userEvents.add(cloudEvent);
              cloudEvents.add(newCloudEvent);
              cloudNodeList.add(data);
              cloudNodePath.add(data['path']);
            } else {
              /// CHECK TIMESTAMP
              if (cloudEvents.isNotEmpty) {
                int userEventIndex = userEvents.indexOf(cloudEvent);
                var listToCompare =
                    cloudEvents.where((event) => event.id_event == cloudEvent);
                if (listToCompare.isNotEmpty) {
                  Event toCompare = listToCompare.first;
                  int cloudEventIndex = cloudEvents.indexOf(toCompare);
                  DateTime? evToCompare = toCompare.last_modified;
                  DateTime? newTime = newCloudEvent.last_modified;
                  //AS IN CLOUD, COMPARE DATETIME
                  if (evToCompare != null && newTime != null) {
                    Duration _diff = newTime.difference(evToCompare);
                    if (_diff.inDays > 0) {
                      userEvents[userEventIndex] = cloudEvent;
                      cloudEvents[cloudEventIndex] = newCloudEvent;
                      cloudNodeList[userEventIndex] = (data);
                      cloudNodePath[userEventIndex] = (data['path']);
                    }
                  }
                }
              }
            }
          }
        } catch (e) {
          log("ERROR GETTING SINGLE USER EVENT");
        }
      }
    } catch (e) {
      log("ERROR GETTING CLOUD USER EVENTS");
    }

    ///ALL ABOUT LISTENER
    toolbox_api.SearchCriteria sc = toolbox_api.SearchCriteria();
    NodeListener stListener = NodeListener();

    stListener.addEventHandler(toolbox_api.EventType.delete, deleteHandler);
    stListener.addEventHandler(toolbox_api.EventType.create, createHandler!);
    stListener.addEventHandler(toolbox_api.EventType.update, updateHandler!);
    stListener.addEventHandler(toolbox_api.EventType.rename, renameHandler!);
    await _storageController.registerChangeListener(stListener, sc);
  }

  @override
  Future<void> renameHandler(EventChange event) async {
    log("RENAME HANDLER");
    String eType = event.type.toValueString();
    log("STORAGE LISTENER EVENT TYPE: $eType");
    toolbox_api.Node? oldNode = event.getOldNode;
    String oldPath = oldNode!.path;
    toolbox_api.Node? newNode = event.getNewNode;
    String tlp = newNode!.visibility.toValueString();
    if (tlp.toLowerCase() != 'black' &&
        newNode.path.startsWith(':Local') == false &&
        newNode.path.startsWith(':Global') == false &&
        newNode.tombstone == false) {
      if (cloudNodePath.contains(oldPath)) {
        int index = cloudNodePath.indexOf(oldPath);
        String cloudId = userEvents[index];
        Event toCheck = await cloud.getSingleUserEvent(_username, cloudId);
        if (tlp.toLowerCase() == 'red' && toCheck.type != 'keyvalue') {
          try {
            Map<dynamic, dynamic> encryptedNode =
                await encryptListenerCloudData(newNode);
            toCheck.content = json.encode(encryptedNode);
          } catch (e) {
            log(e.toString());
            toCheck.content =
                json.encode(await convertNodeToJsonString(newNode));
            log('FAILURE GETTING KEYS RENAME HANDLER');
          }
        } else {
          toCheck.content = json.encode(await convertNodeToJsonString(newNode));
        }
        if (newNode.name.toLowerCase() == _geigerScoreAggNodeName) {
          log("****UPDATE GEIGER SCORE*****");
          toCheck.type = 'aggregate_score';
        } else if (newNode.path.startsWith(':Keys')) {
          toCheck.type = 'keyvalue';
        } else {
          toCheck.type = 'user_object';
        }
        toCheck.owner = _username;
        log("RENAME HANDLER - UPDATE NODE TO THE CLOUD");
        if (toCheck.type != 'keyvalue') {
          if (newNode.path.endsWith(':Enterprise:Users:$_username')) {
            String name = await newNode
                .getValue('name')
                .then((value) => value!.getValue("en").toString());
            await cloud.updateUser(_username, name);
          } else {
            await cloud.updateEvent(_username, toCheck.id_event, toCheck);
          }
          cloudNodePath[index] = newNode.path;
        }
      } else {
        String uuid = await cloud.generateUUID();
        Event toCheck = Event(id_event: uuid, tlp: tlp.toUpperCase());
        toCheck.encoding = 'ascii';
        if (tlp.toLowerCase() == 'red') {
          try {
            Map<dynamic, dynamic> encryptedNode =
                await encryptListenerCloudData(newNode);
            toCheck.content = json.encode(encryptedNode);
          } catch (e) {
            log(e.toString());
            toCheck.content =
                json.encode(await convertNodeToJsonString(newNode));
            log('FAILURE GETTING KEYS RENAME HANDLER');
          }
        } else {
          toCheck.content = json.encode(await convertNodeToJsonString(newNode));
        }
        if (newNode.name.toLowerCase() == _geigerScoreAggNodeName) {
          log("****UPDATE GEIGER SCORE*****");
          toCheck.type = 'aggregate_score';
        } else if (newNode.path.startsWith(':Keys')) {
          toCheck.type = 'keyvalue';
        } else {
          toCheck.type = 'user_object';
        }
        toCheck.owner = _username;
        log("RENAME HANDLER - ADD NODE TO THE CLOUD");
        if (toCheck.type != 'keyvalue') {
          if (newNode.path.endsWith(':Enterprise:Users:$_username')) {
            String name = await newNode
                .getValue('name')
                .then((value) => value!.getValue("en").toString());
            await cloud.updateUser(_username, name);
          } else {
            await cloud.createEvent(_username, toCheck);
          }
          toCheck.last_modified = DateTime.now();
          userEvents.add(uuid);
          cloudEvents.add(toCheck);
          cloudNodeList.add(await convertNodeToJsonString(newNode));
          cloudNodePath.add(newNode.path);
        }
      }
    }
    log("END RENAME HANDLER");
  }

  @override
  Future<void> updateHandler(EventChange event) async {
    log("UPDATE HANDLER");
    String eType = event.type.toValueString();
    log("STORAGE LISTENER EVENT TYPE: $eType");
    toolbox_api.Node? newNode = event.getNewNode;
    String tlp = newNode!.visibility.toValueString();
    String fullPath = newNode.path;
    if (tlp.toLowerCase() != 'black' &&
        newNode.path.startsWith(':Local') == false &&
        newNode.path.startsWith(':Global') == false &&
        newNode.tombstone == false) {
      if (cloudNodePath.contains(fullPath)) {
        int index = cloudNodePath.indexOf(fullPath);
        String cloudId = userEvents[index];
        Event toCheck = await cloud.getSingleUserEvent(_username, cloudId);
        toCheck.encoding = 'ascii';
        if (tlp.toLowerCase() == 'red' && toCheck.type != 'keyvalue') {
          try {
            Map<dynamic, dynamic> encryptedNode =
                await encryptListenerCloudData(newNode);
            toCheck.content = json.encode(encryptedNode);
          } catch (e) {
            log(e.toString());
            toCheck.content =
                json.encode(await convertNodeToJsonString(newNode));
            log('FAILURE GETTING KEYS UPDATE HANDLER');
          }
        } else {
          toCheck.content = json.encode(await convertNodeToJsonString(newNode));
        }
        if (newNode.name.toLowerCase() == _geigerScoreAggNodeName) {
          log("****UPDATE GEIGER SCORE*****");
          toCheck.type = 'aggregate_score';
        } else if (newNode.path.startsWith(':Keys')) {
          toCheck.type = 'keyvalue';
        } else {
          toCheck.type = 'user_object';
        }
        toCheck.owner = _username;
        log("UPDATE HANDLER - UPDATE NODE TO THE CLOUD");
        if (toCheck.type != 'keyvalue') {
          if (newNode.path.endsWith(':Enterprise:Users:$_username')) {
            String name = await newNode
                .getValue('name')
                .then((value) => value!.getValue("en").toString());
            await cloud.updateUser(_username, name);
          } else {
            await cloud.updateEvent(_username, toCheck.id_event, toCheck);
          }
        }
      } else {
        String uuid = await cloud.generateUUID();
        Event toCheck = Event(id_event: uuid, tlp: tlp.toUpperCase());
        toCheck.encoding = 'ascii';
        if (tlp.toLowerCase() == 'red' && toCheck.type != 'keyvalue') {
          try {
            Map<dynamic, dynamic> encryptedNode =
                await encryptListenerCloudData(newNode);
            toCheck.content = json.encode(encryptedNode);
          } catch (e) {
            log(e.toString());
            toCheck.content =
                json.encode(await convertNodeToJsonString(newNode));
            log('FAILURE GETTING KEYS UPDATE HANDLER');
          }
        } else {
          toCheck.content = json.encode(await convertNodeToJsonString(newNode));
        }
        if (newNode.name.toLowerCase() == _geigerScoreAggNodeName) {
          log("****UPDATE GEIGER SCORE*****");
          toCheck.type = 'aggregate_score';
        } else if (newNode.path.startsWith(':Keys')) {
          toCheck.type = 'keyvalue';
        } else {
          toCheck.type = 'user_object';
        }
        toCheck.owner = _username;
        log("CREATE HANDLER - ADD NODE TO THE CLOUD");
        if (toCheck.type != 'keyvalue') {
          if (newNode.path.endsWith(':Enterprise:Users:$_username')) {
            String name = await newNode
                .getValue('name')
                .then((value) => value!.getValue("en").toString());
            await cloud.updateUser(_username, name);
          } else {
            await cloud.createEvent(_username, toCheck);
          }
          toCheck.last_modified = DateTime.now();
          userEvents.add(uuid);
          cloudEvents.add(toCheck);
          cloudNodeList.add(await convertNodeToJsonString(newNode));
          cloudNodePath.add(newNode.path);
        }
      }
    }
    log("END UPDATE HANDLER");
  }

  @override
  Future<void> createHandler(EventChange event) async {
    try {
      log("CREATE HANDLER");
      String eType = event.type.toValueString();
      log("STORAGE LISTENER EVENT TYPE: $eType");

      /// AS IS A NEW NODE - IS NOT IN CLOUD
      String tlp = event.newNode!.visibility.toValueString();
      toolbox_api.Node? newNode = event.getNewNode;
      String uuid = await cloud.generateUUID();
      Event toCheck = Event(id_event: uuid, tlp: tlp.toUpperCase());

      /// BETA TEST - IF PAIRING PLUGIN CREATES - IS NOT REPLICATED
      if (newNode!.owner != _devicesPairing &&
          tlp.toLowerCase() != 'black' &&
          newNode.path.startsWith(':Local') == false &&
          newNode.path.startsWith(':Global') == false &&
          newNode.tombstone == false) {
        toCheck.encoding = 'ascii';
        if (tlp.toLowerCase() == 'red') {
          try {
            Map<dynamic, dynamic> encryptedNode =
                await encryptListenerCloudData(newNode);
            toCheck.content = json.encode(encryptedNode);
          } catch (e) {
            log(e.toString());
            toCheck.content =
                json.encode(await convertNodeToJsonString(newNode));
            log('FAILURE GETTING KEYS CREATE HANDLER');
          }
        } else {
          toCheck.content = json.encode(await convertNodeToJsonString(newNode));
        }
        if (newNode.name.toLowerCase() == _geigerScoreAggNodeName) {
          log("****UPDATE GEIGER SCORE*****");
          toCheck.type = 'aggregate_score';
        } else if (newNode.path.startsWith(':Keys')) {
          toCheck.type = 'keyvalue';
        } else {
          toCheck.type = 'user_object';
        }
        toCheck.owner = _username;
        log("CREATE HANDLER - ADD NODE TO THE CLOUD");
        await cloud.createEvent(_username, toCheck);
        toCheck.last_modified = DateTime.now();
        userEvents.add(uuid);
        cloudEvents.add(toCheck);
        cloudNodePath.add(newNode.path);
        cloudNodeList.add(await convertNodeToJsonString(newNode));
      } else {
        log("NODE MUST NOT BE REPLICATED");
        log("TLP: $tlp");
      }
    } catch (e) {
      log(e.toString());
    }
    log("END CREATE HANDLER");
  }

  @override
  Future<void> deleteHandler(EventChange event) async {
    log("DELETE HANDLER");
    String eType = event.type.toValueString();
    toolbox_api.Node? oldNode = event.oldNode;
    log("STORAGE LISTENER EVENT TYPE: $eType");
    if (eType == 'delete') {
      String tlp = event.oldNode!.visibility.toValueString();
      if (tlp.toLowerCase() != 'black' &&
          oldNode != null &&
          oldNode.path.startsWith(':Local') == false &&
          oldNode.path.startsWith(':Global') == false &&
          oldNode.tombstone == false) {
        if (cloudNodePath.contains(oldNode.path)) {
          int index = cloudNodePath.indexOf(oldNode.path);
          log("DELETE HANDLER - DELETE NODE FROM THE CLOUD");
          await cloud.deleteEvent(_username, userEvents[index]);
          cloudNodePath.remove(cloudNodePath[index]);
          userEvents.remove(userEvents[index]);
          cloudEvents.remove(cloudEvents[index]);
          cloudNodeList.remove(cloudNodeList[index]);
        }
      }
    } else {
      log(eType);
    }
    log("END DELETE HANDLER");
  }

  @override
  Future<void> addSettingsConsent(String username, bool settingsValue) async {
    try {
      toolbox_api.Node consentNode = await getNode(':Local:Consent:$username');
      String consentValue = await consentNode
          .getValue('settingsConsent')
          .then((value) => value!.getValue("en").toString());
      if (consentValue != settingsValue.toString()) {
        consentNode.addOrUpdateValue(toolbox_api.NodeValueImpl(
            'settingsConsent', settingsValue.toString()));
        await _storageController.addOrUpdate(consentNode);
      }
    } catch (e) {
      log("CONSENT NODE NOT FOUND. CREATE CONSENT PARENT NODE");
      toolbox_api.Node consentNode =
          toolbox_api.NodeImpl(':Local:Consent', 'ReplicationController');
      consentNode.addOrUpdateValue(toolbox_api.NodeValueImpl(
          'settingsConsent', settingsValue.toString()));
      await _storageController.addOrUpdate(consentNode);
    }
  }

  Future<Map> decryptCloudData(Event event) async {
    log("DECRYPT CLOUD DATA");
    try {
      Map<dynamic, dynamic> data = getPartialJsonString(event.content!);
      Event key = await cloud.getUserEventKey(_username, data['path']);
      toolbox_api.Node keyNode = convertJsonStringToNode(key.getContent!);
      final hexEncodedKey = await keyNode
          .getValue('key')
          .then((value) => value!.getValue("en").toString());

      /// keyPattern: key=["aes-256-cfb:JWWY+/E5Xppta3AsSIsGrWUOKHmv0w3cbfH5VlsG62Y="]
      //final onlyKey = hexEncodedKey.split(':');
      //final String decodedKey = (onlyKey[1]).toString();
      final String userDecodedKey = await decryptWithUserKey(hexEncodedKey, "");
      final keyVal = Enc.Key.fromBase64(userDecodedKey);
      final iv = Enc.IV.fromLength(16);
      final enc = Enc.Encrypter(Enc.AES(keyVal, mode: Enc.AESMode.cfb64));
      //DECRYPT DATA
      Map<dynamic, dynamic> jsonify = jsonDecode(event.content!);
      log(jsonify['custom_fields']);
      String decrypted = enc
          .decrypt(Enc.Encrypted.fromBase64(jsonify['custom_fields']), iv: iv);
      jsonify['custom_fields'] = decrypted;
      log("********* FOR REVIEW PURPOSES *******");
      log(
          "********* DATA DECRYPTION FROM CLOUD TO DEVICE - FUNCTION decryptCloudData*******");
      log(
          "**************************** KEY VALUE ****************************");
      log("keyVal: " + userDecodedKey);
      log(
          "**************************** END KEY VALUE ****************************");
      log(
          "**************************** DECRYPTED VALUE ****************************");
      log(decrypted);
      log(
          "**************************** END DECRYPTED VALUE ****************************");
      log("********* END FOR REVIEW PURPOSES *******");
      return jsonify;
    } catch (e) {
      log("[EXCEPTION] DECRYPT CLOUD DATA");
      log(e.toString());
      throw ReplicationException(e.toString());
    }
  }

  Future<Map> encryptCloudData(toolbox_api.Node node) async {
    log("[ENCRYPT CLOUD DATA] LOCAL NODE");
    String fullPath = node.path;
    toolbox_api.SearchCriteria criteria = toolbox_api.SearchCriteria(
        searchPath: ':Keys', key: 'path', value: fullPath);
    List<toolbox_api.Node> keyList = await _storageController.search(criteria);
    toolbox_api.Node? keys;
    if (keyList.isNotEmpty) {
      for (var entry in keyList) {
        String comparePath = await entry
            .getValue('path')
            .then((value) => value!.getValue("en").toString());
        if (comparePath == fullPath) {
          keys = entry;
          break;
        }
      }
    } else {
      /// IF A PRE EXISTING KEY DOES NOT EXIST
      /// CREATE NEW ONE AND A NODE
      /// REPLICATE KEY TO CLOUD
      String newUUID = Uuid().v4();
      toolbox_api.Node newKey =
          toolbox_api.NodeImpl(':Keys:$newUUID', _username);
      newKey.addOrUpdateValue(toolbox_api.NodeValueImpl('path', fullPath));
      final nodeKey = Enc.Key.fromSecureRandom(32).base64.toString();
      newKey.addOrUpdateValue(
          toolbox_api.NodeValueImpl('key', 'aes-256-cfb:$nodeKey'));
      toolbox_api.Visibility? checkerVisible =
          toolbox_api.VisibilityExtension.valueOf("amber");
      if (checkerVisible != null) {
        newKey.visibility = checkerVisible;
      }
      await _storageController.add(newKey);

      /// POST KEY EVENT
      String keyUUID = await cloud.generateUUID();
      Event keyEvent = Event(id_event: keyUUID, tlp: "AMBER");
      keyEvent.owner = _username;
      keyEvent.encoding = 'ascii';
      keyEvent.type = 'keyvalue';
      try {
        toolbox_api.Node nodeToEvent = await getNode(':Keys:$newUUID');
        final String userEncryptedKey = await encryptWithUserKey(nodeKey);
        nodeToEvent.addOrUpdateValue(
            toolbox_api.NodeValueImpl('key', userEncryptedKey));
        keyEvent.content =
            json.encode(await convertNodeToJsonString(nodeToEvent));
      } catch (e) {
        log("COULD NOT ENCRYPT WITH USER KEY");
      }
      await cloud.createEvent(_username, keyEvent);
      keys = newKey;

      if (userEvents.contains(keyUUID) == false) {
        userEvents.add(keyUUID);
        cloudEvents.add(keyEvent);
        cloudNodeList.add(json.decode(keyEvent.content!));
        cloudNodePath.add(json.decode(keyEvent.content!)['path']);
      }
    }

    final hexEncodedKey = await keys!
        .getValue('key')
        .then((value) => value!.getValue("en").toString());

    /// keyPattern: key=["aes-256-cfb:JWWY+/E5Xppta3AsSIsGrWUOKHmv0w3cbfH5VlsG62Y="]
    final onlyKey = hexEncodedKey.split(':');
    final String decodedKey = onlyKey[1];
    //log("ENCRYPT WITH USER KEY...");
    //final String userEncryptedKey = await encryptWithUserKey(decodedKey);
    //log("KEY ENCRYPTED WITH USER KEY: "+userEncryptedKey);
    //final keyVal = Enc.Key.fromBase64(userEncryptedKey);
    final keyVal = Enc.Key.fromBase64(decodedKey);
    final iv = Enc.IV.fromLength(16);
    final enc = Enc.Encrypter(Enc.AES(keyVal, mode: Enc.AESMode.cfb64));
    Map convertedNode = await convertNodeToJsonString(node);
    Enc.Encrypted encrypted =
        enc.encrypt(json.encode(convertedNode['custom_fields']), iv: iv);
    convertedNode['custom_fields'] = encrypted.base64;
    return convertedNode;
  }

  Future<Map> encryptListenerCloudData(toolbox_api.Node node) async {
    log("[ENCRYPT LISTENER CLOUD DATA] LOCAL NODE");
    String fullPath = node.path;
    toolbox_api.SearchCriteria criteria = toolbox_api.SearchCriteria(
        searchPath: ':Keys', key: 'path', value: fullPath);
    List<toolbox_api.Node> keyList = await _storageController.search(criteria);
    toolbox_api.Node? keys;
    if (keyList.isNotEmpty) {
      for (var entry in keyList) {
        String comparePath = await entry
            .getValue('path')
            .then((value) => value!.getValue("en").toString());
        if (comparePath == fullPath) {
          keys = entry;
          break;
        }
      }

      final hexEncodedKey = await keys!
          .getValue('key')
          .then((value) => value!.getValue("en").toString());

      /// keyPattern: key=["aes-256-cfb:JWWY+/E5Xppta3AsSIsGrWUOKHmv0w3cbfH5VlsG62Y="]
      final onlyKey = hexEncodedKey.split(':');
      final String decodedKey = onlyKey[1];
      log("[ENCRYPT] KEY USED TO ENCRYPT NODE DATA: " + decodedKey);
      final keyVal = Enc.Key.fromBase64(decodedKey);
      final iv = Enc.IV.fromLength(16);
      final enc = Enc.Encrypter(Enc.AES(keyVal, mode: Enc.AESMode.cfb64));
      Map convertedNode = await convertNodeToJsonString(node);
      Enc.Encrypted encrypted =
          enc.encrypt(json.encode(convertedNode['custom_fields']), iv: iv);
      convertedNode['custom_fields'] = encrypted.base64;
      return convertedNode;
    } else {
      return {};
    }
  }

  Future<Map> encryptNodeToShare(toolbox_api.Node node, String keyPath) async {
    late Map convertedNode;
    //String fullPath = node.path;
    //String identifier = node.name;
    try {
      toolbox_api.Node keyNode = await getNode(keyPath);
      final hexEncodedKey = await keyNode
          .getValue('key')
          .then((value) => value!.getValue("en").toString());
      final onlyKey = hexEncodedKey.split(':');
      final String decodedKey = onlyKey[1];
      log("KEY USED TO ENCRYPT NODE DATA: " + decodedKey);
      final keyVal = Enc.Key.fromBase64(decodedKey);
      final iv = Enc.IV.fromLength(16);
      final enc = Enc.Encrypter(Enc.AES(keyVal, mode: Enc.AESMode.cfb64));
      convertedNode = await convertNodeToJsonString(node);
      Enc.Encrypted encrypted =
          enc.encrypt(json.encode(convertedNode['custom_fields']), iv: iv);
      convertedNode['custom_fields'] = encrypted.base64;
      return convertedNode;
    } catch (e) {
      log("[EXCEPTION] DECRYPT CLOUD DATA");
      throw ReplicationException(e.toString());
    }
  }

  Future<Map> decryptSharedNode(Event event, String keyPath) async {
    try {
      toolbox_api.Node keyNode = await getNode(keyPath);
      final hexEncodedKey = await keyNode
          .getValue('key')
          .then((value) => value!.getValue("en").toString());
      //final onlyKey = hexEncodedKey.split(':');
      //final String decodedKey = onlyKey[1];
      final String userDecodedKey = await decryptWithUserKey(hexEncodedKey, "");
      final keyVal = Enc.Key.fromBase64(userDecodedKey);
      final iv = Enc.IV.fromLength(16);
      final enc = Enc.Encrypter(Enc.AES(keyVal, mode: Enc.AESMode.cfb64));
      //DECRYPT DATA
      Map<dynamic, dynamic> jsonify = jsonDecode(event.content!);
      String decrypted = enc.decrypt(
          Enc.Encrypted.fromBase64(jsonify['custom_fields'].toString()),
          iv: iv);
      jsonify['custom_fields'] = decrypted;
      return jsonify;
    } catch (e) {
      log("[EXCEPTION] DECRYPT CLOUD DATA");
      throw ReplicationException(e.toString());
    }
  }

  Future<Map> decryptCloudNode(
      Event score, Event key, String encryptionKey) async {
    try {
      toolbox_api.Node keyNode = convertJsonStringToNode(key.getContent!);
      final hexEncodedKey = await keyNode
          .getValue('key')
          .then((value) => value!.getValue("en").toString());
      //final onlyKey = hexEncodedKey.split(':');
      //final String decodedKey = onlyKey[1];
      final String userDecodedKey =
          await decryptWithUserKey(hexEncodedKey, encryptionKey);
      final keyVal = Enc.Key.fromBase64(userDecodedKey);
      final iv = Enc.IV.fromLength(16);
      final enc = Enc.Encrypter(Enc.AES(keyVal, mode: Enc.AESMode.cfb64));
      //DECRYPT DATA
      Map<dynamic, dynamic> jsonify = jsonDecode(score.content!);
      String decrypted = enc.decrypt(
          Enc.Encrypted.fromBase64(jsonify['custom_fields'].toString()),
          iv: iv);
      jsonify['custom_fields'] = decrypted;
      log("DECRYPTED NODE IS " + jsonify['custom_fields'].toString());
      return jsonify;
    } catch (e) {
      log("[EXCEPTION] DECRYPT CLOUD DATA");
      throw ReplicationException(e.toString());
    }
  }

  Future<String> encryptWithUserKey(String key) async {
    String keyEncrypted = '';
    try {
      log("[ENCRYPT] KEY BEFORE: " + key);
      toolbox_api.Node keys = await getNode(':Keys');
      String userKey = await keys
          .getValue('publicKey')
          .then((value) => value!.getValue("en").toString());
      final keyVal = Enc.Key.fromBase64(userKey);
      final iv = Enc.IV.fromLength(16);
      final enc = Enc.Encrypter(Enc.AES(keyVal, mode: Enc.AESMode.cfb64));
      Enc.Encrypted encrypted = enc.encrypt(key, iv: iv);
      keyEncrypted = encrypted.base64;
      log("[ENCRYPT] KEY ENCRYPTED WITH USER KEY: " + keyEncrypted);
    } catch (e) {
      log("[ENCRYPT] UNABLE TO ENCRYPT WITH USER KEY");
    }
    return keyEncrypted;
  }

  Future<String> decryptWithUserKey(String key, String userKey) async {
    String keyDecrypted = '';
    try {
      log("[DECRYPT] KEY BEFORE: " + key);
      if (userKey.isEmpty) {
        toolbox_api.Node keys = await getNode(':Keys');
        userKey = await keys
            .getValue('publicKey')
            .then((value) => value!.getValue("en").toString());
      }
      final keyVal = Enc.Key.fromBase64(userKey);
      final iv = Enc.IV.fromLength(16);
      final enc = Enc.Encrypter(Enc.AES(keyVal, mode: Enc.AESMode.cfb64));
      String decrypted = enc.decrypt(Enc.Encrypted.fromBase64(key), iv: iv);
      keyDecrypted = decrypted;
      log("[DECRYPT] KEY DECRYPTED WITH USER KEY: " + keyDecrypted);
    } catch (e) {
      log("[DECRYPT] UNABLE TO DECRYPT WITH USER KEY");
    }
    return keyDecrypted;
  }

  @override
  Future<bool> createPairingStructure(
      String qrUserId, String qrDeviceId, String publicKey, String type) async {
    toolbox_api.Node pairingNode;
    List<dynamic> listScoreCustomFields;
    List<dynamic> listDeviceCustomFields;
    if (type.contains('device')) {
      try {
        pairingNode = await getNode(
            '$_enterpriseUsers:$_username:$_devicesPairing:$qrDeviceId');
      } catch (e) {
        log("DEVICES PAIRING STRUCTURE FOR $qrDeviceId DOES NOT EXIST");
        pairingNode = toolbox_api.NodeImpl(
            '$_enterpriseUsers:$_username:$_devicesPairing:$qrDeviceId',
            _devicesPairing);
      }
      try {
        Event score = await cloud.getUserGeigerScore(qrUserId);
        Event key = await cloud.getUserEventKey(qrUserId,
            ":Users:$qrUserId:1234-1234-1234:data:GeigerScoreAggregate");
        Map<dynamic, dynamic> jsonify =
            await decryptCloudNode(score, key, publicKey);
        score.setContent = jsonEncode(jsonify);
        listScoreCustomFields =
            jsonDecode(jsonDecode(score.content!)['custom_fields']);
        int geigerScoreIndex =
            listScoreCustomFields.indexWhere((f) => f['key'] == 'GEIGER_score');
        int numberMetricsIndex = listScoreCustomFields
            .indexWhere((f) => f['key'] == 'number_metrics');
        int threatsScoreIndex = listScoreCustomFields
            .indexWhere((f) => f['key'] == 'threats_score');
        Event device = await cloud.getUserDeviceInfo(qrUserId, qrDeviceId);
        Event keyDevice =
            await cloud.getUserEventKey(qrUserId, ":Devices:" + qrDeviceId);
        jsonify = await decryptCloudNode(device, keyDevice, publicKey);
        device.setContent = jsonEncode(jsonify);
        listDeviceCustomFields =
            jsonDecode(jsonDecode(device.content!)['custom_fields']);
        int deviceNameIndex =
            listDeviceCustomFields.indexWhere((f) => f['key'] == 'name');
        int deviceTypeIndex =
            listDeviceCustomFields.indexWhere((f) => f['key'] == 'type');
        User userQr = await cloud.getUser(qrUserId);
        pairingNode.addOrUpdateValue(
            toolbox_api.NodeValueImpl('userName', userQr.name ?? ""));
        pairingNode.addOrUpdateValue(toolbox_api.NodeValueImpl(
            'deviceName', listDeviceCustomFields[deviceNameIndex]['value']));
        pairingNode.addOrUpdateValue(toolbox_api.NodeValueImpl(
            'deviceType', listDeviceCustomFields[deviceTypeIndex]['value']));
        pairingNode.addOrUpdateValue(toolbox_api.NodeValueImpl(
            'sharedGeigerScore',
            listScoreCustomFields[geigerScoreIndex]['value']));
        pairingNode.addOrUpdateValue(toolbox_api.NodeValueImpl(
            'sharedThreatsScore',
            listScoreCustomFields[threatsScoreIndex]['value']));
        pairingNode.addOrUpdateValue(toolbox_api.NodeValueImpl(
            'sharedNumberOfMetric',
            listScoreCustomFields[numberMetricsIndex]['value']));
        pairingNode.addOrUpdateValue(toolbox_api.NodeValueImpl(
            'sharedScoreDate', DateTime.now().toString()));
        pairingNode
            .addOrUpdateValue(toolbox_api.NodeValueImpl('userUUID', qrUserId));
        toolbox_api.Visibility? checkerConfig =
            toolbox_api.VisibilityExtension.valueOf("amber");
        if (checkerConfig != null) {
          pairingNode.visibility = checkerConfig;
        }
        await _storageController.addOrUpdate(pairingNode);
        log("DEVICES PAIRING STRUCTURE FOR $qrDeviceId WAS CREATED");
        return true;
      } catch (e) {
        log("PAIRING STRUCTURE FAILED TO GET GEIGER SCORE");
        return false;
      }
      /// EMPLOYEE PAIRING
    } else {
      try {
        pairingNode = await getNode(
            '$_enterpriseUsers:$_username:$_employeesPairing:$qrUserId');
      } catch (e) {
        log("EMPLOYEES PAIRING STRUCTURE FOR $qrUserId DOES NOT EXIST");
        pairingNode = toolbox_api.NodeImpl(
            '$_enterpriseUsers:$_username:$_employeesPairing:$qrUserId',
            _devicesPairing);
      }
      try {
        Event score = await cloud.getUserGeigerScore(qrUserId);
        Event key = await cloud.getUserEventKey(qrUserId,
            ":Users:$qrUserId:1234-1234-1234:data:GeigerScoreAggregate");
        Map<dynamic, dynamic> jsonify =
            await decryptCloudNode(score, key, publicKey);
        score.setContent = jsonEncode(jsonify);
        listScoreCustomFields =
            jsonDecode(jsonDecode(score.content!)['custom_fields']);
        int geigerScoreIndex =
            listScoreCustomFields.indexWhere((f) => f['key'] == 'GEIGER_score');
        int numberMetricsIndex = listScoreCustomFields
            .indexWhere((f) => f['key'] == 'number_metrics');
        int threatsScoreIndex = listScoreCustomFields
            .indexWhere((f) => f['key'] == 'threats_score');
        User userQr = await cloud.getUser(qrUserId);
        pairingNode
            .addValue(toolbox_api.NodeValueImpl('userName', userQr.name ?? ""));
        pairingNode
            .addOrUpdateValue(toolbox_api.NodeValueImpl('userUUID', qrUserId));
        pairingNode.addOrUpdateValue(toolbox_api.NodeValueImpl(
            'sharedGeigerScore',
            listScoreCustomFields[geigerScoreIndex]['value']));
        pairingNode.addOrUpdateValue(toolbox_api.NodeValueImpl(
            'sharedThreatsScore',
            listScoreCustomFields[threatsScoreIndex]['value']));
        pairingNode.addOrUpdateValue(toolbox_api.NodeValueImpl(
            'sharedNumberOfMetric',
            listScoreCustomFields[numberMetricsIndex]['value']));
        pairingNode.addOrUpdateValue(toolbox_api.NodeValueImpl(
            'sharedScoreDate', DateTime.now().toString()));
        toolbox_api.Visibility? checkerConfig =
            toolbox_api.VisibilityExtension.valueOf("amber");
        if (checkerConfig != null) {
          pairingNode.visibility = checkerConfig;
        }
        await _storageController.addOrUpdate(pairingNode);
        log("EMPLOYEES PAIRING STRUCTURE FOR $qrUserId WAS CREATED");
        return true;
      } catch (e) {
        log("PAIRING STRUCTURE FAILED TO GET GEIGER SCORE");
        return false;
      }
    }
  }

  @override
  Future<bool> getSensorData() async {
    log('[GET SENSOR DATA]');
    try {
      await getCurrentDeviceUUID();
      await getCurrentCompany();
      try {
        List<Event> sensorDataEvents =
            await cloud.getEventsByTypesAndOwner(_typesOfSensors, _company);
        log("SENSOR EVENTS " + sensorDataEvents.length.toString());

        if (sensorDataEvents.isNotEmpty) {
          for (Event event in sensorDataEvents) {
            try {
              log("SENSOR DATA EVENT " + event.id_event);
              switch (event.type) {
                case "sensorCloudConnected":
                  {
                    await createCloudConnected(event);
                  }
                  break;
                case "sensorRecommendation":
                  {
                    await createRecommendation(event);
                  }
                  break;
                case "sensorStatus":
                  {
                    await createStatusRecommendation(event);
                  }
                  break;
                default:
                  {
                    log('INVALID CLOUD EVENT TYPE $event.type ?? ""');
                  }
                  break;
              }
            } catch (e) {
              log("ERROR GETTING EVENT WITH ID " +
                  event.id_event +
                  " FOR COMPANY $_company");
            }
          }
          log("[GET SENSOR DATA] Finished OK");
        }
      } catch (e) {
        log("COULD NOT FIND SENSOR DATA");
      }
    } catch (e) {
      log("CURRENT DEVICE NOT FOUND IN DEVICE");
    }
    return true;
  }

  @override
  Future<bool> getTrainingData() async {
    log('[GET TRAINING DATA]');
    try {
      await getCurrentUserUUID();
      try {
        List<Event> trainingDataEvents =
            await cloud.getEventsByTypesAndOwner(_typesOfTrainings, _username);
        log("TRAINING EVENTS " + trainingDataEvents.length.toString());

        if (trainingDataEvents.isNotEmpty) {
          for (Event event in trainingDataEvents) {
            try {
              log("TRAINING DATA EVENT " + event.id_event);
              switch (event.type) {
                case "trainingCloudConnected":
                  {
                    await createCloudConnected(event);
                  }
                  break;
                case "trainingRecommendation":
                  {
                    await createRecommendation(event);
                  }
                  break;
                case "trainingStatus":
                  {
                    await createStatusRecommendation(event);
                  }
                  break;
                default:
                  {
                    log('INVALID CLOUD EVENT TYPE $event.type?? ""');
                  }
                  break;
              }
            } catch (e) {
              log("ERROR GETTING EVENT WITH ID " +
                  event.id_event +
                  " FOR USER $_username");
            }
          }
          log("[GET TRAINING DATA] Finished OK");
        }
      } catch (e) {
        log("COULD NOT FIND TRAINING DATA");
      }
    } catch (e) {
      log("CURRENT USER NOT FOUND IN DEVICE");
    }
    return true;
  }

  @override
  Future<bool> updateSensorData(List<String> types) async {
    log('[UPDATE SENSOR DATA]');
    try {
      await getCurrentDeviceUUID();
      await getCurrentCompany();
      var filter = await getDateDaysBefore(7);
      List<Event> sensorDataEvents =
          await cloud.getEventsByTypesAndOwnerAndAfter(types, _company, filter);
      log("UPDATE SENSOR EVENTS " + sensorDataEvents.length.toString());
      //await cleanUp(sensorDataEvents);

      if (sensorDataEvents.isNotEmpty) {
        for (var event in sensorDataEvents) {
          try {
            log("SENSOR DATA EVENT $event");
            switch (event.type) {
              case "sensorCloudConnected":
                {
                  await updateCloudConnected(event);
                }
                break;
              case "sensorRecommendation":
                {
                  await updateRecommendation(event);
                }
                break;
              case "sensorStatus":
                {
                  await updateStatusRecommendation(event);
                }
                break;
              default:
                {
                  log("INVALID CLOUD EVENT TYPE $event.type");
                }
                break;
            }
          } catch (e) {
            log("ERROR GETTING EVENT WITH ID $event FOR COMPANY $_company");
          }
        }
      }
    } catch (e) {
      log("COULD NOT UPDATE SENSOR DATA");
    }
    return true;
  }

  @override
  Future<bool> updateTrainingData(List<String> types) async {
    log('[UPDATE TRAINING DATA]');
    try {
      await getCurrentUserUUID();
      var filter = await getDateDaysBefore(7);
      try {
        List<Event> trainingDataEvents = await cloud
            .getEventsByTypesAndOwnerAndAfter(types, _username, filter);
        log("UPDATE TRAINING EVENTS " + trainingDataEvents.length.toString());

        if (trainingDataEvents.isNotEmpty) {
          for (var event in trainingDataEvents) {
            try {
              log("TRAINING DATA EVENT $event");
              switch (event.type) {
                case "trainingCloudConnected":
                  {
                    await updateCloudConnected(event);
                  }
                  break;
                case "trainingRecommendation":
                  {
                    await updateRecommendation(event);
                  }
                  break;
                case "trainingStatus":
                  {
                    await updateStatusRecommendation(event);
                  }
                  break;
                default:
                  {
                    log("INVALID CLOUD EVENT TYPE $event.type");
                  }
                  break;
              }
            } catch (e) {
              log("ERROR GETTING EVENT WITH ID $event FOR USER $_username");
            }
          }
          log("[UPDATE TRAINING DATA] Finished OK");
        }
      } catch (e) {
        log("COULD NOT FIND TRAINING DATA");
      }
    } catch (e) {
      log("CURRENT USER NOT FOUND IN DEVICE");
    }
    return true;
  }

  Future<bool> createCloudConnected(Event event) async {
    try {
      CompanyEvent cloudEvent =
          CompanyEvent.fromJson(jsonDecode(event.content!));
      var pluginUUID = cloudEvent.pluginId;
      var metricUUID = event.id_event;
      await createCloudConnectedLocalStorage(pluginUUID, metricUUID, true);
      await addOrUpdateCloudConnected(cloudEvent, pluginUUID, metricUUID, true);
      return true;
    } catch (e) {
      log("[ERROR] CREATE CLOUD CONNECTED");
      return false;
    }
  }

  Future<bool> updateCloudConnected(Event event) async {
    try {
      CompanyEvent cloudEvent =
          CompanyEvent.fromJson(jsonDecode(event.content!));
      var pluginUUID = cloudEvent.pluginId;
      var metricUUID = event.id_event;
      toolbox_api.SearchCriteria sensorCriteria = toolbox_api.SearchCriteria(
          searchPath:
              ':Devices:$_currentDevice:$pluginUUID:data:$_metrics:$metricUUID',
          key: 'name',
          value: cloudEvent.name.toString());
      List<toolbox_api.Node> scProf =
          await _storageController.search(sensorCriteria);
      if (scProf.isNotEmpty) {
        //for(toolbox_api.Node node in scProf){
        await addOrUpdateCloudConnected(
            cloudEvent, pluginUUID, metricUUID, true);
        //}
      } else {
        await createCloudConnected(event);
      }
      return true;
    } catch (e) {
      log("[ERROR] UPDATE CLOUD CONNECTED");
      return false;
    }
  }

  Future<bool> createRecommendation(Event event) async {
    try {
      CloudConnectedToolRecommendation cloudEvent =
          CloudConnectedToolRecommendation.fromJson(jsonDecode(event.content!));
      var pluginUUID = cloudEvent.pluginId;
      var recommendationId = cloudEvent.recommendationId;
      await createCloudConnectedLocalStorage(
          pluginUUID, recommendationId, false);
      await addOrUpdateRecommendation(
          cloudEvent, pluginUUID, recommendationId, false);
      return true;
    } catch (e) {
      log("[ERROR] CREATE CLOUD RECOMMENDATION");
      return false;
    }
  }

  Future<bool> updateRecommendation(Event event) async {
    try {
      CloudConnectedToolRecommendation cloudEvent =
          CloudConnectedToolRecommendation.fromJson(jsonDecode(event.content!));
      var pluginUUID = cloudEvent.pluginId;
      var recommendationId = cloudEvent.recommendationId;
      toolbox_api.SearchCriteria sensorCriteria = toolbox_api.SearchCriteria(
          searchPath:
              ':Devices:$_currentDevice:$pluginUUID:data:$_recommendations:$recommendationId',
          key: 'recommendationId',
          value: cloudEvent.recommendationId.toString());
      List<toolbox_api.Node> scProf =
          await _storageController.search(sensorCriteria);
      if (scProf.isNotEmpty) {
        //for(toolbox_api.Node node in scProf){
        await addOrUpdateRecommendation(
            cloudEvent, pluginUUID, recommendationId, false);
        //}
      }
      return true;
    } catch (e) {
      log("[ERROR] UPDATE CLOUD RECOMMENDATION");
      return false;
    }
  }

  Future<bool> createStatusRecommendation(Event event) async {
    try {
      CloudConnectedToolStatusRecommendation cloudEvent =
          CloudConnectedToolStatusRecommendation.fromJson(
              jsonDecode(event.content!));
      var pluginUUID = cloudEvent.pluginId;
      var recommendationId = cloudEvent.recommendationId;
      await createCloudConnectedLocalStorage(
          pluginUUID, recommendationId, true);
      await addOrUpdateStatusRecommendation(
          cloudEvent, pluginUUID, recommendationId, true);
      return true;
    } catch (e) {
      log("[ERROR] CREATE CLOUD STATUS RECOMMENDATION");
      return false;
    }
  }

  Future<bool> updateStatusRecommendation(Event event) async {
    try {
      CloudConnectedToolStatusRecommendation cloudEvent =
          CloudConnectedToolStatusRecommendation.fromJson(
              jsonDecode(event.content!));
      var pluginUUID = cloudEvent.pluginId;
      var recommendationId = cloudEvent.recommendationId;
      toolbox_api.SearchCriteria sensorCriteria = toolbox_api.SearchCriteria(
          searchPath:
              ':Devices:$_currentDevice:$pluginUUID:data:$_metrics:$recommendationId',
          key: 'recommendationId',
          value: cloudEvent.recommendationId.toString());
      List<toolbox_api.Node> scProf =
          await _storageController.search(sensorCriteria);
      if (scProf.isNotEmpty) {
        //for (toolbox_api.Node node in scProf) {
        await addOrUpdateStatusRecommendation(
            cloudEvent, pluginUUID, recommendationId, true);
        //}
      }
      return true;
    } catch (e) {
      log("[ERROR] UPDATE CLOUD STATUS RECOMMENDATION");
      return false;
    }
  }

  Future<void> addOrUpdateCloudConnected(CompanyEvent cloudEvent,
      String? pluginUUID, String? metricUUID, bool metrics) async {
    var location = metrics ? _metrics : _recommendations;
    toolbox_api.Node node = await getNode(
        ':Devices:$_currentDevice:$pluginUUID:data:$location:$metricUUID');
    node.addOrUpdateValue(
        toolbox_api.NodeValueImpl('name', cloudEvent.name.toString()));
    node.addOrUpdateValue(
        toolbox_api.NodeValueImpl('pluginId', cloudEvent.pluginId.toString()));
    node.addOrUpdateValue(
        toolbox_api.NodeValueImpl('minValue', cloudEvent.minValue.toString()));
    node.addOrUpdateValue(
        toolbox_api.NodeValueImpl('maxValue', cloudEvent.maxValue.toString()));
    node.addOrUpdateValue(toolbox_api.NodeValueImpl(
        'GEIGERvalue', cloudEvent.geigerValue.toString()));
    node.addOrUpdateValue(toolbox_api.NodeValueImpl(
        'valueType', cloudEvent.valueType.toString()));
    node.addOrUpdateValue(
        toolbox_api.NodeValueImpl('type', cloudEvent.type.toString()));
    node.addOrUpdateValue(
        toolbox_api.NodeValueImpl('relation', cloudEvent.relation.toString()));
    node.addOrUpdateValue(toolbox_api.NodeValueImpl(
        'threatsImpact', cloudEvent.threatsImpact.toString()));
    node.addOrUpdateValue(
        toolbox_api.NodeValueImpl('flag', cloudEvent.flag.toString()));
    node.addOrUpdateValue(toolbox_api.NodeValueImpl(
        'description', cloudEvent.description.toString()));
    node.addOrUpdateValue(
        toolbox_api.NodeValueImpl('urgency', cloudEvent.urgency.toString()));
    log("NEW/UPDATE CLOUD CONNECTED STRUCTURE " + node.toString());
    await _storageController.addOrUpdate(node);
  }

  Future<void> addOrUpdateRecommendation(
      CloudConnectedToolRecommendation cloudEvent,
      String? pluginUUID,
      String? metricUUID,
      bool metrics) async {
    var location = metrics ? _metrics : _recommendations;
    toolbox_api.Node node = await getNode(
        ':Devices:$_currentDevice:$pluginUUID:data:$location:$metricUUID');
    node.addOrUpdateValue(
        toolbox_api.NodeValueImpl('pluginId', cloudEvent.pluginId.toString()));
    node.addOrUpdateValue(toolbox_api.NodeValueImpl(
        'pluginName', cloudEvent.pluginName.toString()));
    node.addOrUpdateValue(toolbox_api.NodeValueImpl(
        'recommendationId', cloudEvent.recommendationId.toString()));
    node.addOrUpdateValue(toolbox_api.NodeValueImpl(
        'recommendationType', cloudEvent.recommendationType.toString()));
    node.addOrUpdateValue(toolbox_api.NodeValueImpl(
        'shortDescription', cloudEvent.shortDescription.toString()));
    node.addOrUpdateValue(toolbox_api.NodeValueImpl(
        'longDescription', cloudEvent.longDescription.toString()));
    node.addOrUpdateValue(
        toolbox_api.NodeValueImpl('action', cloudEvent.action.toString()));
    node.addOrUpdateValue(toolbox_api.NodeValueImpl(
        'relatedThreatsWeights', cloudEvent.relatedThreatsWeights.toString()));
    node.addOrUpdateValue(
        toolbox_api.NodeValueImpl('costs', cloudEvent.costs.toString()));
    log("NEW/UPDATE RECOMMENDATION STRUCTURE " + node.toString());
    await _storageController.addOrUpdate(node);
  }

  Future<void> addOrUpdateStatusRecommendation(
      CloudConnectedToolStatusRecommendation cloudEvent,
      String? pluginUUID,
      String? metricUUID,
      bool metrics) async {
    var location = metrics ? _metrics : _recommendations;
    toolbox_api.Node node = await getNode(
        ':Devices:$_currentDevice:$pluginUUID:data:$location:$metricUUID');
    node.addOrUpdateValue(
        toolbox_api.NodeValueImpl('pluginId', cloudEvent.pluginId.toString()));
    node.addOrUpdateValue(toolbox_api.NodeValueImpl(
        'pluginName', cloudEvent.pluginName.toString()));
    node.addOrUpdateValue(
        toolbox_api.NodeValueImpl('name', cloudEvent.name.toString()));
    node.addOrUpdateValue(
        toolbox_api.NodeValueImpl('name', cloudEvent.name.toString()));
    node.addOrUpdateValue(toolbox_api.NodeValueImpl(
        'description', cloudEvent.description.toString()));
    node.addOrUpdateValue(toolbox_api.NodeValueImpl(
        'recommendationId', cloudEvent.recommendationId.toString()));
    node.addOrUpdateValue(
        toolbox_api.NodeValueImpl('minValue', cloudEvent.minValue.toString()));
    node.addOrUpdateValue(
        toolbox_api.NodeValueImpl('maxValue', cloudEvent.maxValue.toString()));
    node.addOrUpdateValue(toolbox_api.NodeValueImpl(
        'valueType', cloudEvent.valueType.toString()));
    node.addOrUpdateValue(
        toolbox_api.NodeValueImpl('flag', cloudEvent.flag.toString()));
    node.addOrUpdateValue(toolbox_api.NodeValueImpl(
        'threatsImpact', cloudEvent.threatsImpact.toString()));
    node.addOrUpdateValue(toolbox_api.NodeValueImpl(
        'geigerValue', cloudEvent.geigerValue.toString()));
    log("NEW/UPDATE STATUS RECOMMENDATION STRUCTURE " + node.toString());
    await _storageController.addOrUpdate(node);
  }

  Future<void> createCloudConnectedLocalStorage(
      String? pluginUUID, String? metricUUID, bool metrics) async {
    if (!await nodeExists(':Devices:$_currentDevice:$pluginUUID')) {
      toolbox_api.Node pluginUUIDNode = toolbox_api.NodeImpl(
          ':Devices:$_currentDevice:$pluginUUID', _replicationAPI);
      toolbox_api.Visibility? checkerData =
          toolbox_api.VisibilityExtension.valueOf(_sensorCloudConnectedTLP);
      if (checkerData != null) {
        pluginUUIDNode.visibility = checkerData;
      }
      await _storageController.add(pluginUUIDNode);
    }
    if (!await nodeExists(':Devices:$_currentDevice:$pluginUUID:data')) {
      toolbox_api.Node dataCloudConnected = toolbox_api.NodeImpl(
          ':Devices:$_currentDevice:$pluginUUID:data', _replicationAPI);
      toolbox_api.Visibility? checkerData =
          toolbox_api.VisibilityExtension.valueOf(_sensorCloudConnectedTLP);
      if (checkerData != null) {
        dataCloudConnected.visibility = checkerData;
      }
      await _storageController.add(dataCloudConnected);
    }
    var location = metrics ? _metrics : _recommendations;
    if (!await nodeExists(
        ':Devices:$_currentDevice:$pluginUUID:data:$location')) {
      toolbox_api.Node metricsCloudConnected = toolbox_api.NodeImpl(
          ':Devices:$_currentDevice:$pluginUUID:data:$location',
          _replicationAPI);
      toolbox_api.Visibility? checkerData =
          toolbox_api.VisibilityExtension.valueOf(_sensorCloudConnectedTLP);
      if (checkerData != null) {
        metricsCloudConnected.visibility = checkerData;
      }
      await _storageController.add(metricsCloudConnected);
    }
    if (metricUUID?.isNotEmpty ?? false) {
      if (!await nodeExists(
          ':Devices:$_currentDevice:$pluginUUID:data:$location:$metricUUID')) {
        toolbox_api.Node sensorNode = toolbox_api.NodeImpl(
            ':Devices:$_currentDevice:$pluginUUID:data:$location:$metricUUID',
            _replicationAPI);
        toolbox_api.Visibility? checkerData =
            toolbox_api.VisibilityExtension.valueOf(_sensorCloudConnectedTLP);
        if (checkerData != null) {
          sensorNode.visibility = checkerData;
        }
        await _storageController.add(sensorNode);
      }
    }
  }

  Future<void> getCurrentUserUUID() async {
    if (usernameInitialized == false) {
      toolbox_api.Node local = await getNode(':Local');
      toolbox_api.NodeValue? currentUser = await local.getValue('currentUser');
      _username = currentUser!.value;
      usernameInitialized = true;
    }
    log("CURRENT USER UUID =======> " + _username);
  }

  Future<void> getCurrentDeviceUUID() async {
    if (currentDeviceInitialized == false) {
      toolbox_api.Node local = await getNode(':Local');
      toolbox_api.NodeValue? currentDevice =
          await local.getValue('currentDevice');
      _currentDevice = currentDevice!.value;
      currentDeviceInitialized = true;
    }
    log("CURRENT DEVICE UUID =======> " + _currentDevice);
  }

  Future<void> getCurrentCompany() async {
    if (companyInitialized == false) {
      try {
        /// TODO Check company name is not empty what should we do?
        toolbox_api.Node local = await getNode(':Enterprise');
        toolbox_api.NodeValue? currentCompany = await local.getValue('name');
        _company = currentCompany!.value.isNotEmpty ? currentCompany.value : '';
      } catch (e) {
        _company = '';
      }
      companyInitialized = true;
    }
    log("CURRENT COMPANY UUID =======> " + _company);
  }

  Future<String> getDateDaysBefore(int daysBefore) async {
    DateTime date = DateTime.now();
    date = date.subtract(Duration(days: daysBefore));
    return DateFormat("yyyy-MM-dd'T'hh:mm:ss.SSSS'Z'").format(date);
  }

  Future<void> cleanUp(List<Event> events) async {
    /*if(events.isNotEmpty){
      for(Event e in events){
        CompanyEvent cloudEvent = CompanyEvent.fromJson(jsonDecode(e.content!));
        cloudEvent.pluginId = 'pluginUUIDRAE';
      }
    }*/
  }

  Future<List<toolbox_api.Node>> getSpecificNodes() async {
    log("GET SPECIFIC NODES");
    List<toolbox_api.Node> nodeList = [];
    toolbox_api.Node node;
    try {
      await getCurrentDeviceUUID();
      await getCurrentUserUUID();
      node = await getNode(':Devices:$_currentDevice');
      nodeList.add(node);
      node = await getNode(':Users:$_username');
      nodeList.add(node);
      node = await getNode(
          ':Users:$_username:1234-1234-1234:data:GeigerScoreAggregate');
      nodeList.add(node);
      node = await getNode(':Enterprise:Users:$_username:devicesPairing');
      nodeList.add(node);
      node = await getNode(':Enterprise:Users:$_username:employeesPairing');
      nodeList.add(node);
    } catch (e) {
      log(e.toString());
    }
    return nodeList;
  }
}
